"""

"""
import openeye as toolkit
from openeye import oechem
import copy
import warnings
from .utils import ANGSROM_2_BOHR, _symbols


def mol_from_json(symbols, connectivity, geometry, permute_xyz=False):
    """
    Generate OEMol from QCSchema molecule specs
    Parameters
    ----------
    inp_molecule: dict
        Must have symbols and connectivity and/or geometry
        Note: If geometry is given, the molecule will have a tag indicating that the goemetry came from QCSchema. This
        will ensure that the order of the atoms and configuration is not change for generation of mapped SMILES and
        isomeric SMILES.

    Returns
    -------
    molecule: OEMol

    """

    molecule = oechem.OEMol()
    for s in symbols:
        molecule.NewAtom(_symbols[s])

    # Add connectivity
    for bond in connectivity:
        a1 = molecule.GetAtom(oechem.OEHasAtomIdx(bond[0]))
        a2 = molecule.GetAtom(oechem.OEHasAtomIdx(bond[1]))
        bond_order = bond[-1]
        if not isinstance(bond_order, int) and not bond_order.is_integer():
            raise ValueError("bond order must be a whole number. A bond order of 1.5 is not allowed")
        molecule.NewBond(a1, a2, int(bond_order))

    # Add geometry
    if molecule.NumAtoms() != geometry.shape[0]/3:
        raise ValueError("Number of atoms in molecule does not match length of position array")

    molecule.SetCoords(oechem.OEFloatArray(geometry))
    molecule.SetDimension(3)

    if not permute_xyz:
        # Add tag that the geometry is from JSON and shouldn't be changed.
        geom_tag = oechem.OEGetTag("json_geometry")
        molecule.SetData(geom_tag, True)
    oechem.OEDetermineConnectivity(molecule)
    oechem.OEFindRingAtomsAndBonds(molecule)
    # No need to perceive Bond Order because the information was added from the connectivity table. Apparently, this
    # function does many different perceptions under the hood and it can add implicit hydrogens to divalent N that should be negatively charged.
    #oechem.OEPerceiveBondOrders(molecule)
    # This seems to add hydrogens that are not in the json
    #oechem.OEAssignImplicitHydrogens(molecule)
    oechem.OEAssignFormalCharges(molecule)
    oechem.OEAssignAromaticFlags(molecule)
    oechem.OEPerceiveChiral(molecule)
    oechem.OE3DToAtomStereo(molecule)
    oechem.OE3DToBondStereo(molecule)

    return molecule


def canonical_order_atoms(molecule, in_place=True):
    """
    Canonical order of atom indices. This ensures the map indices are always in the same order.
    Parameters
    ----------
    molecule: oechem.OEMol
    in_place: bool, optional, default True
        If True, the order of atom indices will happen in place. If False, a copy of the molecule with reordered atom
        indices will be returned.

    Returns
    -------
    molecule: OEMol with canonical ordered atom indices.

    """

    if not in_place:
        molecule = copy.deepcopy(molecule)

    oechem.OECanonicalOrderAtoms(molecule)
    oechem.OECanonicalOrderBonds(molecule)

    vatm = []
    for atom in molecule.GetAtoms():
        if atom.GetAtomicNum() != oechem.OEElemNo_H:
            vatm.append(atom)
    molecule.OrderAtoms(vatm)

    vbnd = []
    for bond in molecule.GetBonds():
        if bond.GetBgn().GetAtomicNum() != oechem.OEElemNo_H and bond.GetEnd().GetAtomicNum() != oechem.OEElemNo_H:
            vbnd.append(bond)
    molecule.OrderBonds(vbnd)

    molecule.Sweep()

    for bond in molecule.GetBonds():
        if bond.GetBgnIdx() > bond.GetEndIdx():
            bond.SwapEnds()

    if not in_place:
        return molecule


def mol_to_smiles(molecule, isomeric=True, explicit_hydrogen=True, mapped=True):
    """
    Generate canonical SMILES with OpenEye.
    Parameters
    ----------

    molecule: oechem.OEMol
    isomeric: bool
        If True, SMILES will include chirality and stereo bonds
    explicit_hydrogen: bool
        If True, SMILES will include explicit hydrogen
    mapped: bool
        If True, will include map indices (In order of OpenEye omega canonical ordering)
    generate_conformer: bool, optional. Default True
        Generating conformer is needed to infer stereochemistry if SMILES does not have stereochemistry specified. Sometimes,
        however, this can be very slow because the molecule has many rotatable bonds. Then it is recommended to turn
        off generate_conformer but the stereochemistry might not be specified in the isomeric SMILES

    Returns
    -------
    smiles str

    """

    molecule = oechem.OEMol(molecule)

    if has_atom_map(molecule):
        remove_atom_map(molecule)

    if explicit_hydrogen:
        if not has_explicit_hydrogen(molecule):
            oechem.OEAddExplicitHydrogens(molecule)

    # First check if geometry from JSON exists
    try:
        JSON_geometry = molecule.GetData('json_geometry')
    except ValueError:
        JSON_geometry = False

    if isomeric:
        if not has_stereo_defined(molecule):
            raise ValueError("Smiles must have stereochemistry defined.")

    if not explicit_hydrogen and not mapped and isomeric:
        return oechem.OEMolToSmiles(molecule)
    if not explicit_hydrogen and not mapped and not isomeric:
        return oechem.OECreateSmiString(molecule, oechem.OESMILESFlag_Canonical | oechem.OESMILESFlag_RGroups)

    if not mapped and explicit_hydrogen and isomeric:
        return oechem.OECreateSmiString(molecule, oechem.OESMILESFlag_Hydrogens | oechem.OESMILESFlag_ISOMERIC)

    if not mapped and explicit_hydrogen and not isomeric:
        return oechem.OECreateSmiString(molecule, oechem.OESMILESFlag_Hydrogens | oechem.OESMILESFlag_Canonical |
                                        oechem.OESMILESFlag_RGroups)

    if not JSON_geometry:
        # canonical order of atoms if input was SMILES or permute_xyz is true
        canonical_order_atoms(molecule)

    for atom in molecule.GetAtoms():
        atom.SetMapIdx(atom.GetIdx() + 1)

    if mapped and not explicit_hydrogen:
        raise Warning("Tagged SMILES must include hydrogens to retain order")

    if mapped and not isomeric:
        raise Warning("Tagged SMILES must include stereochemistry ")

    return oechem.OEMolToSmiles(molecule)

def generate_index(mol, labeled_atoms):
    """
    Generate SMILES with labeled atoms

    Parameters
    ----------
    mol : oechem.OEMol
        Must have atom indices that correspond to labeled atoms
    labeled_atoms : tuple of ints
        ints should be -1 of atom map

    Returns
    -------
    labeled SMILES

    """
    for a in mol.GetAtoms():
        if not a.GetMapIdx()-1 in labeled_atoms:
            a.SetMapIdx(0)
    for i, l in enumerate(labeled_atoms):
        a = mol.GetAtom(oechem.OEHasMapIdx(l+1))
        a.SetMapIdx(i + 1)
    return oechem.OEMolToSmiles(mol)

def get_connectivity_table(molecule, inverse_map):
    """
    Generate connectivity table for molecule using map indices from atom map
    Parameters
    ----------
    molecule: oechem.OEMol
    inverse_map: dict {atom_idx:atom_map}

    Returns
    -------
    connectivity_table: list of lists
        [[atom1, atom2, bond_order]...]

    """
    connectivity_table = [[inverse_map[bond.GetBgnIdx()]-1, inverse_map[bond.GetEndIdx()]-1, bond.GetOrder()] for bond
                          in molecule.GetBonds()]

    return connectivity_table


def get_atom_map(molecule, mapped_smiles, strict=True):
    """
    Map tag in mapped SMILES to atom idx using a substructure search
    A substructure search finds chemically equivalent matches so if atoms are symmetrical, they can flip. The mapped
    SMILES used for the pattern is first used to generate a molecule with the map indices, the order is canonicalized
    and then it is used for the substructure search pattern. This ensures that symmetrical do not flip but there is no
    guarantee that it won't happen.

    Parameters
    ----------
    molecule: oechem.OEMOl
        Must have explicit hydrogen
    mapped_smiles: str
        explicit hydrogen SMILES with map indices on every atom

    Returns
    -------
    atom_map: dict
        {map_idx:atom_idx}

    """
    # check that smiles has explicit hydrogen and map indices
    mapped_mol = oechem.OEMol()
    oechem.OESmilesToMol(mapped_mol, mapped_smiles)
    if not has_atom_map(mapped_mol):
        raise ValueError("Mapped SMILES must have map indices for all atoms and hydrogens")
    # Check molecule for explicit hydrogen
    if not has_explicit_hydrogen(molecule) and strict:
        raise ValueError("Molecule must have explicit hydrogens")

    # canonical order mapped mol to ensure atom map is always generated in the same order
    canonical_order_atoms(mapped_mol)
    aopts = oechem.OEExprOpts_DefaultAtoms
    bopts = oechem.OEExprOpts_DefaultBonds
    ss = oechem.OESubSearch(mapped_mol, aopts, bopts)
    oechem.OEPrepareSearch(molecule, ss)
    ss.SetMaxMatches(1)

    atom_map = {}
    matches = [m for m in ss.Match(molecule)]
    if not matches:
        raise RuntimeError("MCSS failed for {}, smiles: {}".format(oechem.OEMolToSmiles(molecule), mapped_smiles))
    for match in matches:
        for ma in match.GetAtoms():
            atom_map[ma.pattern.GetMapIdx()] = ma.target.GetIdx()

    # sanity check
    mol = oechem.OEGraphMol()
    oechem.OESubsetMol(mol, match, True)
    matched_smiles = mol_to_smiles(mol, isomeric=False, explicit_hydrogen=False, mapped=False)
    molcopy = oechem.OEMol(molecule)
    smiles = mol_to_smiles(molcopy, isomeric=False, explicit_hydrogen=False, mapped=False)
    pattern_smiles = mol_to_smiles(mapped_mol, isomeric=False, explicit_hydrogen=False, mapped=False)
    if not matched_smiles == smiles == pattern_smiles:
        raise RuntimeError("Matched molecule, input molecule and mapped SMILES are not the same ")
    return atom_map


def get_map_ordered_geometry(molecule, atom_map):
    """
    Generate geometry list of size 3*N in the order of the atom map

    Parameters
    ----------
    molecule: OEMol with 1 conformer or OEConf
    atom_map: dict
        map_idx:atom_idx

    Returns
    -------
    symbols: list of str
        list of symbols
    geometry: list of ints
        xyz geometry. List is length N*3
    """

    if not molecule.GetDimension() == 3:
        raise RuntimeError("Molecule must have 3D coordinates for generating a QCSchema molecule")

    if isinstance(molecule, oechem.OEMol):
        if molecule.GetMaxConfIdx() != 1:
            raise Warning("The molecule must have at least and at most 1 conformation")

    symbols = []
    geometry = []
    for mapping in range(1, molecule.NumAtoms()+1):
        idx = atom_map[mapping]
        atom = molecule.GetAtom(oechem.OEHasAtomIdx(idx))
        syb = oechem.OEGetAtomicSymbol(atom.GetAtomicNum())
        symbols.append(syb)
        for i in range(3):
            geometry.append(molecule.GetCoords()[atom.GetIdx()][i]*ANGSROM_2_BOHR)

    return symbols, geometry


def has_explicit_hydrogen(molecule):
    #ToDo test this function well
    """

    Parameters
    ----------
    molecule

    Returns
    -------

    """
    return not oechem.OEHasImplicitHydrogens(molecule)


def add_explicit_hydrogen(molecule):
    oechem.OEAddExplicitHydrogens(molecule)
    return molecule


def get_symbols(molecule):
    return [oechem.OEGetAtomicSymbol(a.GetAtomicNum()) for a in molecule.GetAtoms()]


def has_stereo_defined(molecule):
    """
    Check if any stereochemistry in undefined.
    Parameters
    ----------
    molecule: OEMol

    Returns
    -------
    bool: True if all stereo chemistry is defined.
        If any stereochemsitry is undefined, raise and exception.

    """

    # perceive stereochemistry
    oechem.OEPerceiveChiral(molecule)
    oechem.OE3DToAtomStereo(molecule)
    oechem.OE3DToBondStereo(molecule)

    unspec_chiral = False
    unspec_db = False
    problematic_atoms = list()
    problematic_bonds = list()
    for atom in molecule.GetAtoms():
        if atom.IsChiral() and not atom.HasStereoSpecified(oechem.OEAtomStereo_Tetrahedral):
            # Check if handness is specified
            v = []
            for nbr in atom.GetAtoms():
                v.append(nbr)
            stereo = atom.GetStereo(v, oechem.OEAtomStereo_Tetrahedral)
            if stereo == oechem.OEAtomStereo_Undefined:
                unspec_chiral = True
                problematic_atoms.append((atom.GetIdx(), oechem.OEGetAtomicSymbol(atom.GetAtomicNum())))
    for bond in molecule.GetBonds():
        if bond.IsChiral() and not bond.HasStereoSpecified(oechem.OEBondStereo_CisTrans):
            v = []
            for neigh in bond.GetBgn().GetAtoms():
                if neigh != bond.GetEnd():
                    v.append(neigh)
                    break
            for neigh in bond.GetEnd().GetAtoms():
                if neigh != bond.GetBgn():
                    v.append(neigh)
                    break
            stereo = bond.GetStereo(v, oechem.OEBondStereo_CisTrans)

            if stereo == oechem.OEBondStereo_Undefined:
                unspec_db = True
                a1 = bond.GetBgn()
                a2 = bond.GetEnd()
                a1_idx = a1.GetIdx()
                a2_idx = a2.GetIdx()
                a1_s = oechem.OEGetAtomicSymbol(a1.GetAtomicNum())
                a2_s = oechem.OEGetAtomicSymbol(a2.GetAtomicNum())
                bond_order = bond.GetOrder()
                problematic_bonds.append((a1_idx, a1_s, a2_idx, a2_s, bond_order))
    if unspec_chiral or unspec_db:
        warnings.warn("Stereochemistry is unspecified. Problematic atoms {}, problematic bonds {}, SMILES: {}".format(
                problematic_atoms,
                problematic_bonds, oechem.OEMolToSmiles(molecule)))
        return False
    else:
        return True


def has_atom_map(molecule):
    """
    Checks if any atom has map indices. Will return True if even only one atom has a map index
    Parameters
    ----------
    molecule: oechem.OEMOl

    Returns
    -------
    bool

    """
    IS_MAPPED = False
    for atom in molecule.GetAtoms():
            if atom.GetMapIdx() != 0:
                IS_MAPPED = True
                return IS_MAPPED
    return IS_MAPPED


def is_missing_atom_map(molecule):
    """
    Checks if any atom in molecule is missing a map index. If even only one atom is missing a map index will return True
    Parameters
    ----------
    molecule: oechem.OEMOl

    Returns
    -------
    bool

    """
    MISSING_ATOM_MAP = False
    for atom in molecule.GetAtoms():
            if atom.GetMapIdx() == 0:
                MISSING_ATOM_MAP = True
                return MISSING_ATOM_MAP
    return MISSING_ATOM_MAP


def remove_atom_map(molecule, keep_map_data=True):
    """
    Remove atom map but store it in atom data. In place.
    Parameters
    ----------
    molecule: OEMol with map indices
    keep_map_data: bool, optional, default True
        If True, will store map indices in 'MapIdx' field in atom data

    """
    for atom in molecule.GetAtoms():
        if atom.GetMapIdx() != 0:
            if keep_map_data:
                atom.SetData('MapIdx', atom.GetMapIdx())
            atom.SetMapIdx(0)


def restore_atom_map(molecule):
    """
    Restore atom map from atom data. In place.
    Parameters
    ----------
    molecule: OEMol
        Must have 'MapIdx' in atom data dictionary
    """
    for atom in molecule.GetAtoms():

        if atom.HasData('MapIdx'):
            atom.SetMapIdx(atom.GetData('MapIdx'))

def add_atom_map(molecule, in_place=True):
    """
    Add canonical ordered atom maps to existing molecule
    Parameters
    ----------
    molecule : oechem.OEMol
    in_place : bool, optional, default True

    Returns
    -------
    mapped OEMol if in_place is False

    """
    # First canonical order atoms
    if in_place:
        canonical_order_atoms(molecule, in_place=in_place)
    else:
        molecule = canonical_order_atoms(molecule, in_place=in_place)

    for atom in molecule.GetAtoms():
        atom.SetMapIdx(atom.GetIdx() + 1)
    if not in_place:
        return molecule

def is_map_canonical(molecule):
    """
    Check if map indices on molecule is in canonical order.
    Reorder atoms to be in canonical order and compare the atom order to the map indices.

    Parameters
    ----------
    molecule: OEMol created from the mapped SMILES

    Returns
    -------
    True if canonical ordered atoms match map indices. False otherwise.

    """
    molcopy = oechem.OEMol(molecule)
    # reorder molcopy
    canonical_order_atoms(molcopy, in_place=True)
    # Now check that map indices are +1 on atom indices
    for a in molcopy.GetAtoms():
        if a.GetMapIdx() != a.GetIdx() + 1:
            return False
    return True



"""util functions for test suite"""
from pkg_resources import resource_filename
import os
import re


def get_fn(filename, written=False):
    """Get the full path to one of the reference files shipped for testing

        These files are in torsionfit/testing/reference

    :param
        name: str
            Name of file to load

    :returns
        fn : str
            full path to file
    """
    if written:
        fn = resource_filename('cmiles', os.path.join('tests', 'reference', 'writes', filename))
    else:
        fn = resource_filename('cmiles', os.path.join('tests', 'reference', filename))

    #if not os.path.exists(fn):
    #    raise ValueError('%s does not exist. If you just added it you will have to re install' % fn)

    return fn


def get_smiles_lists(f1, f2):

    smiles = []
    for f in (f1, f2):
        smiles.append(get_smiles_list(f))
    return list(zip(smiles[0], smiles[-1]))


def get_smiles_list(f):
    smiles = open(f, 'r').read().split('\n')[:-1]
    for i in range(len(smiles)):
        smiles[i] = smiles[i].split(' ')[0]
    return smiles


"""
Utility functions for cmiles generator
"""
import numpy as np
import copy
import collections
import warnings

try:
    from rdkit import Chem
    has_rdkit = True
except ImportError:
    has_rdkit = False

try:
    from openeye import oechem
    if not oechem.OEChemIsLicensed():
        has_openeye = False
    has_openeye = True
except ImportError:
    has_openeye = False

if not has_openeye and not has_rdkit:
    raise ImportError("Must have openeye or rdkit installed")

_symbols = {'H': 1, 'He': 2,
            'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10,
            'Na': 11, 'Mg': 12, 'Al': 13,' Si': 14, 'P': 15, 'S': 16, 'Cl': 17, 'Ar': 18,
            'K': 19, 'Ca': 20, 'Sc': 21, 'Ti': 22, 'V': 23, 'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29,
            'Zn': 30, 'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39, 'Zr': 40,
            'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50, 'Sb': 51,
            'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60, 'Pm': 61, 'Sm': 62,
            'Eu': 63,' Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70, 'Lu': 71, 'Hf': 72, 'Ta': 73,
            'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80, 'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po':84,
            'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90,' Pa': 91, 'U': 92, 'Np': 93, 'Pu': 94, 'Am': 95,
            'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100, 'Md': 101, 'No': 102, 'Lr': 103, 'Rf': 104, 'Db': 105,
            'Sg': 106, 'Bh': 107, 'Hs': 108, 'Mt': 109}
BOHR_2_ANGSTROM = 0.529177210
ANGSROM_2_BOHR = 1. / BOHR_2_ANGSTROM


def load_molecule(inp_molecule, toolkit='openeye', **kwargs):
    """
    Load molecule.

    Input is restrictive. Allowed inputs are:

    1. Isomeric SMILES
    2. JSON serialized molecule

    Parameters
    ----------
    inp_molecule: str or dict
        isomeric SMILES or QCSChema
    toolkit: str, optional, default openeye.
        cheminformatics toolkit to use

    Returns
    -------
    molecule:
        `oechem.OEMOl` or `rdkit.Chem.Mol`
    """

    # Check input
    if isinstance(inp_molecule, dict):
        # This is a JSON molecule.
        molecule = mol_from_json(inp_molecule, toolkit=toolkit, **kwargs)

    elif isinstance(inp_molecule, str):
        if toolkit == 'openeye' and has_openeye:
            molecule = oechem.OEMol()
            if not oechem.OESmilesToMol(molecule, inp_molecule):
                raise ValueError("The supplied SMILES {} could not be parsed".format(inp_molecule))
        elif toolkit == 'rdkit' and has_rdkit:
            a = Chem.rdmolfiles.SmilesParserParams()
            a.removeHs = False
            molecule = Chem.MolFromSmiles(inp_molecule, a)
            if not molecule:
                raise ValueError("The supplied SMILES {} could not be parsed".format(inp_molecule))
        else:
            raise ValueError("Only openeye and rdkit toolkits are supported")
    else:
        raise ValueError("Only QCSchema serialized molecule or an isomric SMILES are valid inputs")

    return molecule


def mol_from_json(inp_molecule, toolkit='openeye', **kwargs):
    """
    Load a molecule from QCSchema

    see `QCSchema <https://molssi-qc-schema.readthedocs.io/en/latest/index.html#>`_

     Required fields for the QCSchema molecule:

    1. symbols
    2. geometry
    3. connectivity

    Parameters
    ----------
    inp_molecule: dict
       QCSchema molecule with `symbols`, `geometry` and `connectivity`
    toolkit: str, optional. Default openeye
        cheminformatics toolkit to use. Currently supports `openeye` and `rdkit`
    **permute_xyz: bool, optional, default False
        If False, will add flag to molecule such that the mapped SMILES retains the order of serialized geometry. If True,
        mapped SMILES will be in canonical order and serialized geometry will have to be reordered.

    Returns
    -------
    molecule
      `oechem.OEMol` or `rdkit.Chem.Mol`

    """
    # Check fields
    required_fields = ['symbols', 'geometry', 'connectivity']
    for key in required_fields:
        if key not in inp_molecule:
            raise KeyError("input molecule must have {}".format(key))

    symbols = inp_molecule['symbols']
    connectivity = inp_molecule['connectivity']

    # convert to Angstrom.
    geometry = np.asarray(inp_molecule['geometry'], dtype=float)*BOHR_2_ANGSTROM
    if len(symbols) != geometry.shape[0]/3:
        raise ValueError("Number of atoms in molecule does not match length of position array")

    if toolkit == 'openeye' and has_openeye:
        import cmiles._cmiles_oe as mol_toolkit
    elif toolkit == 'rdkit' and has_rdkit:
        import cmiles._cmiles_rd as mol_toolkit
    else:
        raise ValueError("Only openeye and rdkit backends are supported")

    molecule = mol_toolkit.mol_from_json(symbols, connectivity, geometry, **kwargs)

    return molecule


def mol_to_smiles(molecule, **kwargs):
    """
    Generate canonical smiles from molecule

    Parameters
    ----------
    molecule:
        `oechem.OEMol` or `rdkit.Chem.Mol`
    **isomeric: bool, optional, default True
        If False, SMILES will not include stereo information
    **explicit_hydrogen: bool, optional, default True
        If True, SMILES will have explicit hydrogen.
    **mapped: bool, optional, default True
        If True, SMILES will have map indices

        Example: O=O will be ``[O:1]=[O:2]``


    Returns
    -------
    str
        SMILES

    """
    molecule = copy.deepcopy(molecule)
    toolkit = _set_toolkit(molecule)
    if has_atom_map(molecule):
        remove_atom_map(molecule)
    return toolkit.mol_to_smiles(molecule, **kwargs)

def to_canonical_label(mapped_smiles, labeled_atoms, toolkit='openeye'):
    """
    Generate human readable index with labeled torsions, angles, or bonds

    Parameters
    ----------
    mapped_smiles : str
        SMILES with map indices
    labeled_atoms : tuple of int
        ints should correspond to map indices -1 in mapped SMILES

    Returns
    -------
    labeled SMILES

    """

    mol = load_molecule(mapped_smiles, toolkit=toolkit)
    toolkit = _set_toolkit(mol)
    if not has_atom_map(mol):
        raise RuntimeError("SMILES must have map indices")
    return toolkit.generate_index(mol, labeled_atoms)


def mol_to_hill_molecular_formula(molecule):
    """
    Generate Hill sorted empirical formula.

    Hill sorted first lists C and H and then all other symbols in alphabetical
    order

    Parameters
    ----------
    molecule:
    `oechem.OEMol` or `rdkit.Chem.Mol`

    Returns
    -------
    str
        hill sorted empirical formula
    """

    # check molecule
    toolkit = _set_toolkit(molecule)
    if not has_explicit_hydrogen(molecule):
        molecule = toolkit.add_explicit_hydrogen(molecule)
    symbols = toolkit.get_symbols(molecule)

    count = collections.Counter(x.title() for x in symbols)

    hill_sorted = []
    for k in ['C', 'H']:
        # remove C and H from count
        if k in count:
            c = count.pop(k)
            hill_sorted.append(k)
            if c > 1:
                hill_sorted.append(str(c))

    for k in sorted(count.keys()):
        c = count[k]
        hill_sorted.append(k)
        if c > 1:
            hill_sorted.append(str(c))

    return "".join(hill_sorted)


def mol_to_map_ordered_qcschema(molecule, mapped_smiles, multiplicity=1, **kwargs):
    """
    Genereate JSON serialize following `QCSchema specs <https://molssi-qc-schema.readthedocs.io/en/latest/index.html#>`_

    Geometry, symbols and connectivity table ordered according to map indices in mapped SMILES

    Parameters
    ----------
    molecule:
        `oechem.OEMol` or `rdkit.Chem.Mol`
        **molecuel must have a conformer**.
    molecule_ids: dict
        cmiles generated molecular ids.
    multiplicity: int, optional, defualt 1
        multiplicity of molecule

    Returns
    -------
    dict
        JSON serialized molecule following QCSchema specs

    """
    toolkit = _set_toolkit(molecule)
    atom_map = toolkit.get_atom_map(molecule, mapped_smiles, **kwargs)

    connectivity = get_connectivity_table(molecule, atom_map)
    symbols, geometry = toolkit.get_map_ordered_geometry(molecule, atom_map)
    charge = get_charge(molecule)

    qcschema_mol = {'symbols': symbols, 'geometry': geometry, 'connectivity': connectivity,
                    'molecular_charge': charge, 'molecular_multiplicity': multiplicity}

    return qcschema_mol


def get_atom_map(molecule, mapped_smiles, **kwargs):
    """
    Get mapping of map index -> atom index

    Parameters
    ----------
    molecule:
        `oechem.OEMol` or `rdkit.Chem.Mol`
    mapped_smiles: str
        explicit hydrogen mapped SMILES

    Returns
    -------
    atom_map: dict
        dictionary mapping `{map_index: atom_index}`

    """
    toolkit = _set_toolkit(molecule)

    atom_map = toolkit.get_atom_map(molecule, mapped_smiles)
    return atom_map


def get_connectivity_table(molecule, atom_map):
    """
    Generate connectivity table

    Parameters
    ----------
    molecule:
        oechem.Mol or rdkit.Chem.Mol
    atom_map: dict
        ``{map_idx : atom_idx}``

    Returns
    -------
    list: list of lists
        lists of atoms bonded and the bond order
        [[map_idx_1, map_idx_2, bond_order] ...]

    """

    toolkit = _set_toolkit(molecule)
    inverse_map = dict(zip(atom_map.values(), atom_map.keys()))
    return toolkit.get_connectivity_table(molecule, inverse_map)


def permute_qcschema(json_mol, molecule_ids, **kwargs):
    """
    permute geometry and symbols to correspond to map indices on mapped SMILES

    Parameters
    ----------
    json_mol: dict
        JSON serialized molecule.

        Required fields: `symbols`, `geometry`, `connectivity` and `multiplicity`
    molecule_ids: dict
        cmiles generated molecular ids

    Returns
    -------
   dict
        JSON serialized molecule. `symbols`, `geometry`, and `connectivity` ordered according to map indices on mapped
        SMILES.

        Also includes `identifiers` field with cmiles generated identifiers.

    """
    molecule = mol_from_json(json_mol, **kwargs)
    ordered_qcschema = mol_to_map_ordered_qcschema(molecule, molecule_ids, json_mol['molecular_multiplicity'])

    return ordered_qcschema


def has_atom_map(molecule):
    """
    Check if molecule has atom map indices. Will return True even if only one atom has map index

    Parameters
    ----------
    molecule:
        `oechem.Mol` or `rdkit.Chem.Mol`

    Returns
    -------
    bool
        True if has one map index. False if molecule has no map indices

    """
    toolkit = _set_toolkit(molecule)
    return toolkit.has_atom_map(molecule)


def is_missing_atom_map(molecule):
    """
    Check if any atom in molecule is missing atom map index

    Parameters
    ----------
    molecule:
        oechem.Mol or rdkit.Chem.Mol

    Returns
    -------
    bool
        True if even if only one atom map is missing. False if all atoms have atom maps.

    """
    toolkit = _set_toolkit(molecule)
    return toolkit.is_missing_atom_map(molecule)


def is_map_canonical(molecule):
    """
    Check if map indices on molecule is in caononical order

    Parameters
    ----------
    molecule:
        `oechem.Mol` or `rdkit.Chem.Mol`

    Returns
    -------
    bool

    """
    toolkit = _set_toolkit(molecule)
    return toolkit.is_map_canonical(molecule)


def remove_atom_map(molecule, **kwargs):
    """
    Remove atom map from molecule

    Parameters
    ----------
    molecule
        `oechem.OEMol` or `rdkit.Chem.Mol`
    keep_map_data: bool, optional, default True
        If True, will save map indices in atom data

    """
    toolkit = _set_toolkit(molecule)
    toolkit.remove_atom_map(molecule, **kwargs)


def restore_atom_map(molecule):
    """
    Restore atom map from atom data in place

    Parameters
    ----------
    molecule
        `oechem.OEMol` or `rdkit.Chem.Mol`

    """
    toolkit = _set_toolkit(molecule)
    toolkit.restore_atom_map(molecule)
    if not has_atom_map(molecule):
        warnings.warn("There were no atom maps in atom data to restore")

def add_atom_map(molecule, **kwargs):
    """
    Add canonical ordered atom map to molecule
    Parameters
    ----------
    molecule :
        `oechem.OEMOl` or `rdkit.Chem.Mol`

    Returns
    -------
    molecule with map indices

    """
    toolkit = _set_toolkit(molecule)
    return toolkit.add_atom_map(molecule, **kwargs)


def has_stereo_defined(molecule):
    """
    Checks if molecule has all stereo defined.

    Parameters
    ----------
    molecule:
        `oechem.OEMol` or `rdkit.Chem.Mol`

    Returns
    -------
    bool
        True if all stereo defined, False otherwise

    Notes
    -----
    This does not check if all chirality or bond stereo are consistent. The best way to check is to try to generate a
    3D conformer. If stereo information is inconsistent, this will fail.
    """
    toolkit = _set_toolkit(molecule)
    return toolkit.has_stereo_defined(molecule)


def has_explicit_hydrogen(molecule):
    #ToDo: Use option in RDKit to generate explicit hydrogen molecules from explicit hydrogen SMILES
    """
    Check if molecule has explicit hydrogen.

    Parameters
    ----------
    molecule:
        `oechem.OEMol` or `rdkit.Chem.Mol`

    Returns
    -------
    bool
        True if has all explicit H. False otherwise.

    """
    toolkit = _set_toolkit(molecule)
    return toolkit.has_explicit_hydrogen(molecule)


def add_explicit_hydrogen(molecule):
    """
    Add explicit hydrogen to molecule

    Parameters
    ----------
    molecule:
        `oechem.OEMol` or `rdkit.Chem.Mol`

    Returns
    -------
    molecule
        `oechem.OEMol` or `rdkit.Chem.Mol` with explict hydrogen

    """
    toolkit = _set_toolkit(molecule)
    return toolkit.add_explicit_hydrogen(molecule)


def get_charge(molecule):
    """
    Get charge state of molecule

    Parameters
    ----------
    molecule:
        `oechem.OEMol` or `rdkit.Chem.Mol`

    Returns
    -------
    int
        total charge of molecule

    """

    charge = 0
    for atom in molecule.GetAtoms():
        charge += atom.GetFormalCharge()
    return charge


def _set_toolkit(molecule):
    """
    Set toolkit to use by checking molecule instance and if the toolkit is installed

    Parameters
    ----------
    molecule:
        oechem.OEMol or rdkit.Chem.Mol

    Returns
    -------
    toolkit: module
        either cmiles._cmiles_oe or cmiles._cmiles_rd
    """

    if has_openeye and isinstance(molecule, oechem.OEMolBase):
        import cmiles._cmiles_oe as toolkit
    elif has_rdkit and isinstance(molecule, Chem.rdchem.Mol):
        import cmiles._cmiles_rd as toolkit
    else:
        raise RuntimeError("Must have openeye or rdkit installed")
    return toolkit


def invert_atom_map(atom_map):
    """
    Invert atom map `{map_idx:atom_idx} --> {atom_idx:map_idx}`

    Parameters
    ----------
    atom_map: dict
        `{map_idx:atom_idx}`

    Returns
    -------
    dict
        `{atom_idx:map_idx}`

    """
    return dict(zip(atom_map.values(), atom_map.keys()))



"""
Unit and regression test for the cmiles package.
"""

# Import package, test suite, and other packages as needed

import pytest
import sys
import numpy as np
from .utils import get_fn, get_smiles_lists
import cmiles

toolkits = list()
toolkits_name = list()
if cmiles.utils.has_rdkit:
    from cmiles import _cmiles_rd
    from rdkit import Chem

    toolkits.append(_cmiles_rd)
    toolkits_name.append('rdkit')

if cmiles.utils.has_openeye:
    from cmiles import _cmiles_oe
    from openeye import oechem
    toolkits.append(_cmiles_oe)
    toolkits_name.append('openeye')

using_rdkit = pytest.mark.skipif(not cmiles.utils.has_rdkit, reason="Cannot run without RDKit")
using_openeye = pytest.mark.skipif(not cmiles.utils.has_openeye, reason="Cannot run without OpenEye")


def test_cmiles_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "cmiles" in sys.modules

@pytest.fixture
def smiles_input():
    return([
        "C[C@@H](c1c(ccc(c1Cl)F)Cl)OC",
        "CN(C)C/C=C/C(=O)NC1=C(C=C2C(=C1)C(=NC=N2)NC3=CC(=C(C=C3)F)Cl)O[C@H]4CCOC4",
        "N[C@@](C)(F)C(=O)O",
        "c1ccc2c(c1)C=CCC2=O",
        "c1ccc2c(c1)cccc2O",
        "CC[N+]#C",
        "Cc1ccccc1",
        "CC1=C(C(=O)C2=C(C1=O)[N@@]3C[C@H]4[C@@H]([C@@]3([C@@H]2COC(=O)N)OC)N4)N",
        "C1[C@H]2N1C(=O)N=C2N",
        "[H][C@@]1(C)C[N@@]1C1=CC(=O)C2=C(C(CO)=C(N2C)C2=CC=CC=C2)C1=O"
    ])

@pytest.fixture
def rd_iso_expected():
    return(["CO[C@@H](C)c1c(Cl)ccc(F)c1Cl",
            "CN(C)C/C=C/C(=O)Nc1cc2c(Nc3ccc(F)c(Cl)c3)ncnc2cc1O[C@H]1CCOC1",
            "C[C@](N)(F)C(=O)O",
            "O=C1CC=Cc2ccccc21",
            "Oc1cccc2ccccc12",
            "C#[N+]CC",
            "Cc1ccccc1"

    ])

@pytest.fixture
def rd_map_expected():
    return (['[F:1][c:7]1[c:5]([H:14])[c:6]([H:15])[c:8]([Cl:2])[c:10]([C@@:13]([O:4][C:11]([H:16])([H:17])[H:18])([C:12]([H:19])([H:20])[H:21])[H:22])[c:9]1[Cl:3]',
             '[O:1]=[C:8](/[C:9](=[C:10](/[C:30]([N:27]([C:28]([H:45])([H:46])[H:47])[C:29]([H:48])([H:49])[H:50])([H:51])[H:52])[H:36])[H:35])[N:25]([c:18]1[c:17]([O:7][C@:34]2([H:59])[C:32]([H:55])([H:56])[O:6][C:31]([H:53])([H:54])[C:33]2([H:57])[H:58])[c:20]([H:41])[c:23]2[n:5][c:11]([H:37])[n:4][c:21]([N:26]([c:19]3[c:13]([H:39])[c:12]([H:38])[c:15]([F:2])[c:16]([Cl:3])[c:14]3[H:40])[H:44])[c:24]2[c:22]1[H:42])[H:43]',
             '[O:1]=[C:4]([O:3][H:8])[C@:7]([F:2])([N:5]([H:9])[H:10])[C:6]([H:11])([H:12])[H:13]',
             '[O:1]=[C:2]1[c:9]2[c:7]([H:16])[c:5]([H:14])[c:6]([H:15])[c:8]([H:17])[c:10]2[C:3]([H:12])=[C:4]([H:13])[C:11]1([H:18])[H:19]',
             '[O:1]([c:9]1[c:5]([H:16])[c:3]([H:14])[c:7]([H:18])[c:10]2[c:6]([H:17])[c:2]([H:13])[c:4]([H:15])[c:8]([H:19])[c:11]12)[H:12]',
             '[C:1](#[N+:2][C:4]([C:3]([H:6])([H:7])[H:8])([H:9])[H:10])[H:5]',
             '[c:1]1([H:8])[c:2]([H:9])[c:4]([H:11])[c:6]([C:7]([H:13])([H:14])[H:15])[c:5]([H:12])[c:3]1[H:10]'
            ])

@pytest.fixture
def rd_can_expected():
    return (['COC(C)c1c(Cl)ccc(F)c1Cl',
             'CN(C)CC=CC(=O)Nc1cc2c(Nc3ccc(F)c(Cl)c3)ncnc2cc1OC1CCOC1',
             'CC(N)(F)C(=O)O',
             'O=C1CC=Cc2ccccc21',
             'Oc1cccc2ccccc12',
             'C#[N+]CC',
             'Cc1ccccc1'
    ])

@pytest.fixture
def rd_h_expected():
    return (['[H][c]1[c]([H])[c]([Cl])[c]([C@@]([H])([O][C]([H])([H])[H])[C]([H])([H])[H])[c]([Cl])[c]1[F]',
             '[H]/[C]([C](=[O])[N]([H])[c]1[c]([O][C@]2([H])[C]([H])([H])[O][C]([H])([H])[C]2([H])[H])[c]([H])[c]2[n][c]([H])[n][c]([N]([H])[c]3[c]([H])[c]([H])[c]([F])[c]([Cl])[c]3[H])[c]2[c]1[H])=[C](/[H])[C]([H])([H])[N]([C]([H])([H])[H])[C]([H])([H])[H]',
             '[H][O][C](=[O])[C@]([F])([N]([H])[H])[C]([H])([H])[H]',
             '[H][C]1=[C]([H])[C]([H])([H])[C](=[O])[c]2[c]([H])[c]([H])[c]([H])[c]([H])[c]21',
             '[H][O][c]1[c]([H])[c]([H])[c]([H])[c]2[c]([H])[c]([H])[c]([H])[c]([H])[c]12',
             '[H][C]#[N+][C]([H])([H])[C]([H])([H])[H]',
             '[H][c]1[c]([H])[c]([H])[c]([C]([H])([H])[H])[c]([H])[c]1[H]'
    ])
@pytest.fixture
def oe_iso_expected():
    return(['C[C@@H](c1c(ccc(c1Cl)F)Cl)OC',
            'CN(C)C/C=C/C(=O)Nc1cc2c(cc1O[C@H]3CCOC3)ncnc2Nc4ccc(c(c4)Cl)F',
            'C[C@](C(=O)O)(N)F',
            'c1ccc2c(c1)C=CCC2=O',
            'c1ccc2c(c1)cccc2O',
            'CC[N+]#C',
            'Cc1ccccc1'
    ])

@pytest.fixture
def oe_map_expected():
    return (['[H:14][c:1]1[c:2]([c:5]([c:3]([c:6]([c:4]1[F:11])[Cl:13])[C@:9]([H:22])([C:7]([H:16])([H:17])[H:18])[O:10][C:8]([H:19])([H:20])[H:21])[Cl:12])[H:15]',
             '[H:35][c:1]1[c:2]([c:12]([c:13]([c:5]([c:9]1[N:27]([H:58])[c:14]2[c:7]3[c:3]([c:10]([c:11]([c:4]([c:8]3[n:25][c:6]([n:26]2)[H:40])[H:38])[O:32][C@:21]4([C:18]([C:19]([O:31][C:20]4([H:47])[H:48])([H:45])[H:46])([H:43])[H:44])[H:49])[N:28]([H:59])[C:17](=[O:30])/[C:15](=[C:16](\\[H:42])/[C:24]([H:56])([H:57])[N:29]([C:22]([H:50])([H:51])[H:52])[C:23]([H:53])([H:54])[H:55])/[H:41])[H:37])[H:39])[Cl:34])[F:33])[H:36]',
             '[H:8][C:2]([H:9])([H:10])[C@:3]([C:1](=[O:5])[O:6][H:13])([N:4]([H:11])[H:12])[F:7]',
             '[H:12][c:1]1[c:2]([c:4]([c:6]2[c:5]([c:3]1[H:14])[C:7](=[C:8]([C:10]([C:9]2=[O:11])([H:18])[H:19])[H:17])[H:16])[H:15])[H:13]',
             '[H:12][c:1]1[c:2]([c:5]([c:9]2[c:8]([c:4]1[H:15])[c:6]([c:3]([c:7]([c:10]2[O:11][H:19])[H:18])[H:14])[H:17])[H:16])[H:13]',
             '[H:5][C:1]#[N+:4][C:3]([H:9])([H:10])[C:2]([H:6])([H:7])[H:8]',
             '[H:8][c:1]1[c:2]([c:4]([c:6]([c:5]([c:3]1[H:10])[H:12])[C:7]([H:13])([H:14])[H:15])[H:11])[H:9]',
             '[H:27][C@:9]1([C:1]2=[C:3]([C:6](=[O:21])[C:2](=[C:4]([C:5]2=[O:20])[N:18]([H:39])[H:40])[C:13]([H:30])([H:31])[H:32])[N@:17]3[C@@:12]1([C@@:11]4([C@:10]([C:8]3([H:25])[H:26])([N:16]4[H:38])[H:28])[H:29])[O:24][C:14]([H:33])([H:34])[H:35])[C:15]([H:36])([H:37])[O:23][C:7](=[O:22])[N:19]([H:41])[H:42]',
             '[H:11][C@@:4]12[C:1](=[N:5][C:2](=[O:8])[N:6]1[C:3]2([H:9])[H:10])[N:7]([H:12])[H:13]',
             '[H:25][c:1]1[c:2]([c:4]([c:6]([c:5]([c:3]1[H:27])[H:29])[c:9]2[c:8]([c:7]3[c:10]([n:20]2[C:18]([H:37])([H:38])[H:39])[C:13](=[O:23])[C:11](=[C:14]([C:12]3=[O:22])[N@@:21]4[C@:16]([C:15]4([H:31])[H:32])([H:33])[C:17]([H:34])([H:35])[H:36])[H:30])[C:19]([H:40])([H:41])[O:24][H:42])[H:28])[H:26]'])

@pytest.fixture
def oe_can_expected():
    return (['CC(c1c(ccc(c1Cl)F)Cl)OC',
             'CN(C)CC=CC(=O)Nc1cc2c(cc1OC3CCOC3)ncnc2Nc4ccc(c(c4)Cl)F',
             'CC(C(=O)O)(N)F',
             'c1ccc2c(c1)C=CCC2=O',
             'c1ccc2c(c1)cccc2O',
             'CC[N+]#C',
             'Cc1ccccc1'
    ])

@pytest.fixture
def oe_h_expected():
    return (['[H]c1c(c(c(c(c1F)Cl)[C@]([H])(C([H])([H])[H])OC([H])([H])[H])Cl)[H]',
             '[H]c1c(c(c(c(c1N([H])c2c3c(c(c(c(c3nc(n2)[H])[H])O[C@]4(C(C(OC4([H])[H])([H])[H])([H])[H])[H])N([H])C(=O)/C(=C(\\[H])/C([H])([H])N(C([H])([H])[H])C([H])([H])[H])/[H])[H])[H])Cl)F)[H]',
             '[H]C([H])([H])[C@](C(=O)O[H])(N([H])[H])F',
             '[H]c1c(c(c2c(c1[H])C(=C(C(C2=O)([H])[H])[H])[H])[H])[H]',
             '[H]c1c(c(c2c(c1[H])c(c(c(c2O[H])[H])[H])[H])[H])[H]',
             '[H]C#[N+]C([H])([H])C([H])([H])[H]',
             '[H]c1c(c(c(c(c1[H])[H])C([H])([H])[H])[H])[H]'])


@using_rdkit
def test_rdkit_isomeric(smiles_input, rd_iso_expected):
    """testing rdkit isomeric canonical smiles"""
    for i, o in zip(smiles_input, rd_iso_expected):
        rd_mol = Chem.MolFromSmiles(i)
        assert cmiles.utils.mol_to_smiles(rd_mol, isomeric=True, mapped=False, explicit_hydrogen=False) == o


@using_rdkit
def test_rdkit_map(smiles_input, rd_map_expected):
    """Testing rdkit canonical map ordering"""
    for i, o in zip(smiles_input, rd_map_expected):
        rd_mol = Chem.MolFromSmiles(i)
        assert cmiles.utils.mol_to_smiles(rd_mol, isomeric=True, mapped=True, explicit_hydrogen=True) == o


@using_rdkit
def test_rdkit_canonical(smiles_input, rd_can_expected):
    """Testing rdkit canonical smiles"""
    for i, o in zip(smiles_input, rd_can_expected):
        rd_mol = Chem.MolFromSmiles(i)
        assert cmiles.utils.mol_to_smiles(rd_mol, isomeric=False, mapped=False, explicit_hydrogen=False) == o


@using_rdkit
def test_rdkit_explicit_h(smiles_input, rd_h_expected):
    """Testing rdkit explicit hydrogen"""
    for i, o in zip(smiles_input, rd_h_expected):
        rd_mol = Chem.MolFromSmiles(i)
        assert cmiles.utils.mol_to_smiles(rd_mol, isomeric=True, mapped=False, explicit_hydrogen=True) == o


@using_openeye
def test_openeye_explicit_h(smiles_input, oe_h_expected):
    """Testing openeye explicit hydrogen"""
    for i, o in zip(smiles_input, oe_h_expected):
        oe_mol = oechem.OEMol()
        oechem.OESmilesToMol(oe_mol, i)
        assert cmiles.utils.mol_to_smiles(oe_mol, isomeric=True, mapped=False, explicit_hydrogen=True) == o


@using_openeye
def test_openeye_isomeric(smiles_input, oe_iso_expected):
    """testing rdkit isomeric canonical smiles"""
    for i, o in zip(smiles_input, oe_iso_expected):
        oe_mol = oechem.OEMol()
        oechem.OESmilesToMol(oe_mol, i)
        assert cmiles.utils.mol_to_smiles(oe_mol, isomeric=True, mapped=False, explicit_hydrogen=False) == o


@using_openeye
def test_openeye_map(smiles_input, oe_map_expected):
    """Testing rdkit canonical map ordering"""
    for i, o in zip(smiles_input, oe_map_expected):
        oe_mol = oechem.OEMol()
        oechem.OESmilesToMol(oe_mol, i)
        assert cmiles.utils.mol_to_smiles(oe_mol, isomeric=True, mapped=True, explicit_hydrogen=True) == o


@using_openeye
def test_openeye_canonical(smiles_input, oe_can_expected):
    """Testing rdkit canonical smiles"""
    for i, o in zip(smiles_input, oe_can_expected):
        oe_mol = oechem.OEMol()
        oechem.OESmilesToMol(oe_mol, i)
        assert cmiles.utils.mol_to_smiles(oe_mol, isomeric=False, mapped=False, explicit_hydrogen=False) == o


@using_rdkit
@using_openeye
def test_oe_cmiles():
    """Regression test oe cmiles"""
    expected_output = {'canonical_smiles': 'CN(C)CC=CC(=O)Nc1cc2c(cc1OC3CCOC3)ncnc2Nc4ccc(c(c4)Cl)F',
                       'canonical_isomeric_smiles': 'CN(C)C/C=C/C(=O)Nc1cc2c(cc1O[C@@H]3CCOC3)ncnc2Nc4ccc(c(c4)Cl)F',
                       'canonical_isomeric_explicit_hydrogen_smiles':
                           '[H]c1c(c(c(c(c1N([H])c2c3c(c(c(c(c3nc(n2)[H])[H])O[C@@]4(C(C(OC4([H])[H])([H])[H])([H])[H])[H])N([H])C(=O)/C(=C(\\[H])/C([H])([H])N(C([H])([H])[H])C([H])([H])[H])/[H])[H])[H])Cl)F)[H]',
                       'canonical_explicit_hydrogen_smiles':
                           '[H]c1c(c(c(c(c1N([H])c2c3c(c(c(c(c3nc(n2)[H])[H])OC4(C(C(OC4([H])[H])([H])[H])([H])[H])[H])N([H])C(=O)C(=C([H])C([H])([H])N(C([H])([H])[H])C([H])([H])[H])[H])[H])[H])Cl)F)[H]',
                       'canonical_isomeric_explicit_hydrogen_mapped_smiles':
                           '[H:35][c:1]1[c:2]([c:12]([c:13]([c:5]([c:9]1[N:27]([H:58])[c:14]2[c:7]3[c:3]([c:10]([c:11]([c:4]([c:8]3[n:25][c:6]([n:26]2)[H:40])[H:38])[O:32][C@@:21]4([C:18]([C:19]([O:31][C:20]4([H:47])[H:48])([H:45])[H:46])([H:43])[H:44])[H:49])[N:28]([H:59])[C:17](=[O:30])/[C:15](=[C:16](\\[H:42])/[C:24]([H:56])([H:57])[N:29]([C:22]([H:50])([H:51])[H:52])[C:23]([H:53])([H:54])[H:55])/[H:41])[H:37])[H:39])[Cl:34])[F:33])[H:36]',
                       'provenance': 'cmiles_0.0.0+1.geb7d850.dirty_openeye_2018.Feb.b6'}
    smiles = 'CN(C)C/C=C/C(=O)Nc1cc2c(cc1O[C@@H]3CCOC3)ncnc2Nc4ccc(c(c4)Cl)F'
    output = cmiles.get_molecule_ids(smiles, toolkit='openeye', strict=False)
    assert expected_output['canonical_smiles'] == output['canonical_smiles']
    assert expected_output['canonical_isomeric_smiles'] == output['canonical_isomeric_smiles']
    assert expected_output['canonical_isomeric_explicit_hydrogen_smiles'] == output['canonical_isomeric_explicit_hydrogen_smiles']
    assert expected_output['canonical_explicit_hydrogen_smiles'] == output['canonical_explicit_hydrogen_smiles']
    assert expected_output['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == output['canonical_isomeric_explicit_hydrogen_mapped_smiles']


@using_rdkit
def test_rd_cmiles():
    """Regression test rdkit cmiles"""
    expected_output = {'canonical_smiles': 'CN(C)CC=CC(=O)Nc1cc2c(Nc3ccc(F)c(Cl)c3)ncnc2cc1OC1CCOC1',
 'canonical_isomeric_smiles': 'CN(C)C/C=C/C(=O)Nc1cc2c(Nc3ccc(F)c(Cl)c3)ncnc2cc1O[C@H]1CCOC1',
 'canonical_isomeric_explicit_hydrogen_smiles': '[H]/[C]([C](=[O])[N]([H])[c]1[c]([O][C@]2([H])[C]([H])([H])[O][C]([H])([H])[C]2([H])[H])[c]([H])[c]2[n][c]([H])[n][c]([N]([H])[c]3[c]([H])[c]([H])[c]([F])[c]([Cl])[c]3[H])[c]2[c]1[H])=[C](/[H])[C]([H])([H])[N]([C]([H])([H])[H])[C]([H])([H])[H]',
 'canonical_explicit_hydrogen_smiles': '[H][C]([C](=[O])[N]([H])[c]1[c]([O][C]2([H])[C]([H])([H])[O][C]([H])([H])[C]2([H])[H])[c]([H])[c]2[n][c]([H])[n][c]([N]([H])[c]3[c]([H])[c]([H])[c]([F])[c]([Cl])[c]3[H])[c]2[c]1[H])=[C]([H])[C]([H])([H])[N]([C]([H])([H])[H])[C]([H])([H])[H]',
 'canonical_isomeric_explicit_hydrogen_mapped_smiles': '[O:1]=[C:8](/[C:9](=[C:10](/[C:30]([N:27]([C:28]([H:45])([H:46])[H:47])[C:29]([H:48])([H:49])[H:50])([H:51])[H:52])[H:36])[H:35])[N:25]([c:18]1[c:17]([O:7][C@:34]2([H:59])[C:32]([H:55])([H:56])[O:6][C:31]([H:53])([H:54])[C:33]2([H:57])[H:58])[c:20]([H:41])[c:23]2[n:5][c:11]([H:37])[n:4][c:21]([N:26]([c:19]3[c:13]([H:39])[c:12]([H:38])[c:15]([F:2])[c:16]([Cl:3])[c:14]3[H:40])[H:44])[c:24]2[c:22]1[H:42])[H:43]',
 'provenance': 'cmiles_0.0.0+7.gc71f3a6.dirty_rdkit_2018.03.3'}
    smiles = 'CN(C)C/C=C/C(=O)Nc1cc2c(Nc3ccc(F)c(Cl)c3)ncnc2cc1O[C@H]1CCOC1'
    output = cmiles.get_molecule_ids(smiles, toolkit='rdkit', strict=False)
    assert expected_output['canonical_smiles'] == output['canonical_smiles']
    assert expected_output['canonical_isomeric_smiles'] == output['canonical_isomeric_smiles']
    assert expected_output['canonical_isomeric_explicit_hydrogen_smiles'] == output['canonical_isomeric_explicit_hydrogen_smiles']
    assert expected_output['canonical_explicit_hydrogen_smiles'] == output['canonical_explicit_hydrogen_smiles']
    assert expected_output['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == output['canonical_isomeric_explicit_hydrogen_mapped_smiles']


@using_openeye
@using_rdkit
def test_diff_smiles():
    """Test different SMILES of same molecule"""
    input_smiles = ['C[C@H](c1c(ccc(c1Cl)F)Cl)OC', 'CO[C@H](C)c1c(Cl)ccc(F)c1Cl']
    cmiles_1 = cmiles.get_molecule_ids(input_smiles[0], strict=False)
    cmiles_2 = cmiles.get_molecule_ids(input_smiles[1], strict=False)
    assert cmiles_1['canonical_smiles'] == cmiles_2['canonical_smiles']
    assert cmiles_1['canonical_isomeric_smiles'] == cmiles_2['canonical_isomeric_smiles']
    assert cmiles_1['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == cmiles_2['canonical_isomeric_explicit_hydrogen_mapped_smiles']

    cmiles_1 = cmiles.get_molecule_ids(input_smiles[0], toolkit='rdkit', strict=False)
    cmiles_2 = cmiles.get_molecule_ids(input_smiles[1], toolkit='rdkit', strict=False)
    assert cmiles_1['canonical_smiles'] == cmiles_2['canonical_smiles']
    assert cmiles_1['canonical_isomeric_smiles'] == cmiles_2['canonical_isomeric_smiles']
    assert cmiles_1['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == cmiles_2['canonical_isomeric_explicit_hydrogen_mapped_smiles']

@using_openeye
@using_rdkit
def test_initial_iso():
    """test given chirality"""

    input_smiles = ["CC[C@@H](C)N",  "CC[C@H](C)N"]
    cmiles_1 = cmiles.get_molecule_ids(input_smiles[0], strict=False)
    cmiles_2 = cmiles.get_molecule_ids(input_smiles[-1], strict=False)
    assert cmiles_1['canonical_smiles'] == cmiles_2['canonical_smiles']
    assert cmiles_1['canonical_isomeric_smiles'] != cmiles_2['canonical_isomeric_smiles']

    cmiles_1 = cmiles.get_molecule_ids(input_smiles[0], toolkit='rdkit', strict=False)
    cmiles_2 = cmiles.get_molecule_ids(input_smiles[-1], toolkit='rdkit', strict=False)
    assert cmiles_1['canonical_smiles'] == cmiles_2['canonical_smiles']
    assert cmiles_1['canonical_isomeric_smiles'] != cmiles_2['canonical_isomeric_smiles']


@using_rdkit
@pytest.mark.parametrize("input, output", get_smiles_lists(get_fn('drug_bank_stereo.smi'), get_fn('drug_bank_mapped_smi_rd_2020.09.1.smi')))
def test_drug_bank_rd(input, output):
    """

    Parameters
    ----------
    input
    output

    Returns
    -------

    """

    mol = Chem.MolFromSmiles(input)
    assert cmiles.utils.mol_to_smiles(mol, mapped=True, isomeric=True, explicit_hydrogen=True) == output


@using_openeye
@pytest.mark.parametrize("input, output", get_smiles_lists(get_fn('drug_bank_stereo.smi'), get_fn('drug_bank_mapped_smi_oe.smi')))
def test_drug_bank_oe(input, output):
    """

    Parameters
    ----------
    input
    output

    Returns
    -------

    """
    mol = oechem.OEMol()
    oechem.OEParseSmiles(mol, input)
    assert cmiles.utils.mol_to_smiles(mol, mapped=True, isomeric=True, explicit_hydrogen=True) == output


@using_rdkit
@pytest.mark.parametrize("input, output", get_smiles_lists(get_fn('drug_bank_stereo.smi'), get_fn('drug_bank_inchi_rd_2020.09.1.txt')))
def test_inchi(input, output):
    """Check that inchis are the same"""

    rd_mol = Chem.MolFromSmiles(input)
    rd_inchi = cmiles.generator.get_inchi_and_key(rd_mol)[0]

    assert rd_inchi == output


@using_rdkit
@pytest.mark.parametrize("input, output", get_smiles_lists(get_fn('drug_bank_stereo.smi'), get_fn('drug_bank_inchikey_rd_2020.09.1.txt')))
def test_inchi_key(input, output):
    """Check that inchi key is the same"""

    rd_mol = Chem.MolFromSmiles(input)
    rd_inchi_key = cmiles.generator.get_inchi_and_key(rd_mol)[1]

    assert rd_inchi_key == output

@using_openeye
@using_rdkit
def test_input_mapped():
    smiles = '[H:3][C:1]([H:4])([H:5])[C:2]([H:6])([H:7])[H:8]'
    mol_id = cmiles.get_molecule_ids(smiles)

    mol_1 = cmiles.utils.load_molecule(mol_id['canonical_isomeric_smiles'])
    mol_2 = cmiles.utils.load_molecule(mol_id['canonical_isomeric_explicit_hydrogen_mapped_smiles'])
    assert cmiles.utils.has_atom_map(mol_1) == False
    assert cmiles.utils.has_atom_map(mol_2) == True

    mol_id = cmiles.get_molecule_ids(smiles, toolkit='rdkit', strict=False)

    mol_1 = cmiles.utils.load_molecule(mol_id['canonical_isomeric_smiles'], backend='rdkit')
    mol_2 = cmiles.utils.load_molecule(mol_id['canonical_isomeric_explicit_hydrogen_mapped_smiles'], backend='rdkit')
    assert cmiles.utils.has_atom_map(mol_1) == False
    assert cmiles.utils.has_atom_map(mol_2) == True


def _map_from_json(hooh, backend, map_smiles):

    molecule = cmiles.utils.load_molecule(hooh, toolkit=backend)
    if backend == 'openeye':
        mapped_smiles_1 = cmiles.utils.mol_to_smiles(molecule, isomeric=True, explicit_hydrogen=True, mapped=True)
        molecule.SetData("json_geometry", False)
        mapped_smiles_2 = cmiles.utils.mol_to_smiles(molecule, isomeric=True, explicit_hydrogen=True, mapped=True)
    if backend == 'rdkit':
        mapped_smiles_1 = cmiles.utils.mol_to_smiles(molecule, isomeric=True, explicit_hydrogen=True, mapped=True)
        molecule.SetProp("_json_geometry", '0')
        mapped_smiles_2 = cmiles.utils.mol_to_smiles(molecule, isomeric=True, explicit_hydrogen=True, mapped=True)

    assert mapped_smiles_1 == '[H:1][O:2][O:3][H:4]'
    assert mapped_smiles_2 == map_smiles

@using_openeye
@using_rdkit
@pytest.mark.parametrize("backend, map_smiles", [('openeye', '[H:3][O:1][O:2][H:4]'),
                                                 ('rdkit', '[O:1]([O:2][H:4])[H:3]')])
def test_map_order(backend, map_smiles):
    hooh = {
    'symbols': ['H', 'O', 'O', 'H'],
    'geometry': [
         1.84719633,  1.47046223,  0.80987166,
         1.3126021,  -0.13023157, -0.0513322,
        -1.31320906,  0.13130216, -0.05020593,
        -1.83756335, -1.48745318,  0.80161212
    ],
    'name': 'HOOH',
    'connectivity': [[0, 1, 1], [1, 2, 1], [2, 3, 1]],
    }

    _map_from_json(hooh=hooh, backend=backend, map_smiles=map_smiles)


@using_openeye
@using_rdkit
def test_keep_chiral_stereo():
    """Test that reading from json molecule retains the order of json geometry and stereochemistry"""

    json_mol = {'symbols': ['C', 'C', 'N', 'O', 'F', 'H', 'H', 'H', 'H', 'H', 'H'],
                'geometry': [1.490934395068127, -0.022852472359013117, -1.935709059338355,
                             -0.07992863034848685, -0.42027454585371643, 0.4300901370510521,
                             -1.6008431326210255,  1.7962788702240675,  0.9893952378782299,
                             -1.578310435156546,  -2.623152319435938, 0.12587101271275358,
                              1.5081897367264838, -0.8595839767115931, 2.4023274238804375,
                              2.643487029125874,  -1.686714912858618,  -2.3700985298604698,
                              0.29985967115960716, 0.42241227312506313, -3.568237727722486,
                              2.7917672897488948,  1.5663042901906687, -1.6694857028577224,
                             -0.4416762043595982,  3.317083889862761,  1.2129328698056736,
                             -2.732926456425621,   2.1997415241410825, -0.5153340816908529,
                             -2.648885919666481,  -2.3294408246718734, -1.337378806095166],
                'molecular_charge': 0,
                'molecular_multiplicity': 1,
                'connectivity': [[0, 1, 1],
                  [1, 2, 1],
                  [1, 3, 1],
                  [1, 4, 1],
                  [0, 5, 1],
                  [0, 6, 1],
                  [0, 7, 1],
                  [2, 8, 1],
                  [2, 9, 1],
                  [3, 10, 1]]
                }

    mol_id = cmiles.get_molecule_ids(json_mol)

    assert mol_id['canonical_smiles'] == 'CC(N)(O)F'
    assert mol_id['canonical_isomeric_smiles'] == 'C[C@@](N)(O)F'
    assert mol_id['canonical_isomeric_explicit_hydrogen_smiles'] == '[H]C([H])([H])[C@@](N([H])[H])(O[H])F'
    assert mol_id['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[H:6][C:1]([H:7])([H:8])[C@@:2]([N:3]([H:9])[H:10])([O:4][H:11])[F:5]'

    # generate rd canonicalized smiles - the order should still be as before even though that is not the rdkit canonical
    # order. We want to retain the order for json molecules to their geometry
    mol_id = cmiles.get_molecule_ids(json_mol, toolkit='rdkit')
    assert mol_id['canonical_smiles'] == 'CC(N)(O)F'
    assert mol_id['canonical_isomeric_smiles'] == 'C[C@@](N)(O)F'
    assert mol_id['canonical_isomeric_explicit_hydrogen_smiles'] == '[H][O][C@]([F])([N]([H])[H])[C]([H])([H])[H]'
    assert mol_id['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[C:1]([C@@:2]([N:3]([H:9])[H:10])([O:4][H:11])[F:5])([H:6])([H:7])[H:8]'

    # Now the other stereoisomer
    json_mol = {'symbols': ['C', 'C', 'N', 'O', 'F', 'H', 'H', 'H', 'H', 'H', 'H'],
                'geometry': [1.490934395068127, -0.022852472359013117, -1.935709059338355,
                             -0.07992863034848685, -0.42027454585371643, 0.4300901370510521,
                             -1.6008431326210255, 1.7962788702240675, 0.9893952378782299,
                              1.544484919393002,  -1.0715460728389934, 2.461713642916755,
                             -1.6405346423022924, -2.4261921600567007, 0.04846706513552157,
                              2.643487029125874,  -1.686714912858618,  -2.3700985298604698,
                              0.29985967115960716, 0.42241227312506313, -3.568237727722486,
                              2.7917672897488948,  1.5663042901906687, -1.6694857028577224,
                              -0.4416762043595982, 3.317083889862761, 1.2129328698056736,
                               -2.732926456425621, 2.1997415241410825, -0.5153340816908529,
                                2.4021616230193055, -2.619467530461027, 1.9699541458951846],
                'molecular_charge': 0,
                'molecular_multiplicity': 1,
                'connectivity': [[0, 1, 1],
                [1, 2, 1],
                [1, 3, 1],
                [1, 4, 1],
                [0, 5, 1],
                [0, 6, 1],
                [0, 7, 1],
                [2, 8, 1],
                [2, 9, 1],
                [3, 10, 1]],
                }
    mol_id = cmiles.get_molecule_ids(json_mol)

    assert mol_id['canonical_smiles'] == 'CC(N)(O)F'
    assert mol_id['canonical_isomeric_smiles'] == 'C[C@](N)(O)F'
    assert mol_id['canonical_isomeric_explicit_hydrogen_smiles'] == '[H]C([H])([H])[C@](N([H])[H])(O[H])F'
    assert mol_id['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[H:6][C:1]([H:7])([H:8])[C@:2]([N:3]([H:9])[H:10])([O:4][H:11])[F:5]'

    # generate rd canonicalized smiles - the order should still be as before even though that is not the rdkit canonical
    # order. We want to retain the order for json molecules to their geometry
    mol_id = cmiles.get_molecule_ids(json_mol, toolkit='rdkit')
    assert mol_id['canonical_smiles'] == 'CC(N)(O)F'
    assert mol_id['canonical_isomeric_smiles'] == 'C[C@](N)(O)F'
    assert mol_id['canonical_isomeric_explicit_hydrogen_smiles'] == '[H][O][C@@]([F])([N]([H])[H])[C]([H])([H])[H]'
    assert mol_id['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[C:1]([C@:2]([N:3]([H:9])[H:10])([O:4][H:11])[F:5])([H:6])([H:7])[H:8]'

@using_openeye
@using_rdkit
def test_bond_stereo():
    """Test bond steroe from json molecule"""
    json_mol_from_oe_map = {'symbols': ['C', 'C', 'F', 'Cl', 'H', 'H'],
                'geometry': [0.7558174176630313,
                              -0.9436196701031863,
                              -0.5135812331847833,
                              -0.7123369866046005,
                              0.7689600644555532,
                              0.5875385545305212,
                              1.2485802802219408,
                              -3.180729126504143,
                              0.5903747404566769,
                              -1.3805989906253051,
                              3.6349234648338813,
                              -0.7522673418877901,
                              1.6921967038297914,
                              -0.786834118158881,
                              -2.319716469002742,
                              -1.6036583681666305,
                              0.5072991602038667,
                              2.4076517490881173],
                'molecular_charge': 0,
                'molecular_multiplicity': 1,
                'connectivity': [[0, 1, 2], [0, 2, 1], [1, 3, 1], [0, 4, 1], [1, 5, 1]]}

    json_mol_from_rd_map = {'symbols': ['H', 'H', 'F', 'Cl', 'C', 'C'],
                            'geometry': [1.6921967038297914,
                            -0.786834118158881,
                            -2.319716469002742,
                            -1.6036583681666305,
                            0.5072991602038667,
                            2.4076517490881173,
                            1.2485802802219408,
                            -3.180729126504143,
                            0.5903747404566769,
                            -1.3805989906253051,
                            3.6349234648338813,
                            -0.7522673418877901,
                            0.7558174176630313,
                            -0.9436196701031863,
                            -0.5135812331847833,
                            -0.7123369866046005,
                            0.7689600644555532,
                            0.5875385545305212],
                            'molecular_charge': 0,
                            'molecular_multiplicity': 1,
                            'connectivity': [[4, 5, 2], [4, 2, 1], [5, 3, 1], [4, 0, 1], [5, 1, 1]],}

    id_oe_to_oe = cmiles.get_molecule_ids(json_mol_from_oe_map, toolkit='openeye')
    id_oe_to_rd = cmiles.get_molecule_ids(json_mol_from_oe_map, toolkit='rdkit')
    id_rd_to_oe = cmiles.get_molecule_ids(json_mol_from_rd_map, toolkit='openeye')
    id_rd_to_rd = cmiles.get_molecule_ids(json_mol_from_rd_map, toolkit='rdkit')

    assert id_oe_to_oe['canonical_smiles'] == id_rd_to_oe['canonical_smiles'] == 'C(=CCl)F'
    assert id_oe_to_oe['canonical_isomeric_smiles'] == id_rd_to_oe['canonical_isomeric_smiles'] == 'C(=C/Cl)\\F'
    assert id_oe_to_oe['canonical_explicit_hydrogen_smiles'] == id_rd_to_oe['canonical_explicit_hydrogen_smiles'] == '[H]C(=C([H])Cl)F'
    assert id_oe_to_oe['canonical_isomeric_explicit_hydrogen_smiles'] == id_rd_to_oe['canonical_isomeric_explicit_hydrogen_smiles'] == '[H]/C(=C(/[H])\\Cl)/F'
    assert id_oe_to_oe['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[H:5]/[C:1](=[C:2](/[H:6])\\[Cl:4])/[F:3]'
    assert id_rd_to_oe['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[H:1]/[C:5](=[C:6](/[H:2])\\[Cl:4])/[F:3]'

    assert id_oe_to_rd['canonical_smiles'] == id_rd_to_rd['canonical_smiles'] == 'FC=CCl'
    assert id_oe_to_rd['canonical_isomeric_smiles'] == id_rd_to_rd['canonical_isomeric_smiles'] == 'F/C=C/Cl'
    assert id_oe_to_rd['canonical_explicit_hydrogen_smiles'] == id_rd_to_rd['canonical_explicit_hydrogen_smiles'] == '[H][C]([F])=[C]([H])[Cl]'
    assert id_oe_to_rd['canonical_isomeric_explicit_hydrogen_smiles'] == id_rd_to_rd['canonical_isomeric_explicit_hydrogen_smiles'] == '[H]/[C]([F])=[C](/[H])[Cl]'
    assert id_oe_to_rd['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[C:1](=[C:2](/[Cl:4])[H:6])(\\[F:3])[H:5]'
    assert id_rd_to_rd['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[H:1]/[C:5]([F:3])=[C:6](/[H:2])[Cl:4]'

    # Now the other stereoisomer
    json_mol_from_oe_map = {'symbols': ['C', 'C', 'F', 'Cl', 'H', 'H'],
                 'geometry': [0.7558174176630313,
                  -0.9436196701031863,
                  -0.5135812331847833,
                  -0.7123369866046005,
                  0.7689600644555532,
                  0.5875385545305212,
                  1.2485802802219408,
                  -3.180729126504143,
                  0.5903747404566769,
                  -2.098028327659392,
                  0.3039736895037833,
                  3.4718143896933764,
                  1.6921967038297914,
                  -0.786834118158881,
                  -2.319716469002742,
                  -1.1578923904597032,
                  2.5868975732566115,
                  -0.2324103574431031],
                 'molecular_charge': 0,
                 'molecular_multiplicity': 1,
                 'connectivity': [[0, 1, 2], [0, 2, 1], [1, 3, 1], [0, 4, 1], [1, 5, 1]]
                }
    json_mol_from_rd_map = {'symbols': ['H', 'H', 'F', 'Cl', 'C', 'C'],
                            'geometry': [1.6921967038297914,
                            -0.786834118158881,
                            -2.319716469002742,
                            -1.1578923904597032,
                            2.5868975732566115,
                            -0.2324103574431031,
                            1.2485802802219408,
                            -3.180729126504143,
                            0.5903747404566769,
                            -2.098028327659392,
                            0.3039736895037833,
                            3.4718143896933764,
                            0.7558174176630313,
                            -0.9436196701031863,
                            -0.5135812331847833,
                            -0.7123369866046005,
                            0.7689600644555532,
                            0.5875385545305212],
                            'molecular_charge': 0,
                            'molecular_multiplicity': 1,
                            'connectivity': [[4, 5, 2], [4, 2, 1], [5, 3, 1], [4, 0, 1], [5, 1, 1]]}

    id_oe_to_oe = cmiles.get_molecule_ids(json_mol_from_oe_map, toolkit='openeye')
    id_oe_to_rd = cmiles.get_molecule_ids(json_mol_from_oe_map, toolkit='rdkit')
    id_rd_to_oe = cmiles.get_molecule_ids(json_mol_from_rd_map, toolkit='openeye')
    id_rd_to_rd = cmiles.get_molecule_ids(json_mol_from_rd_map, toolkit='rdkit')

    assert id_oe_to_oe['canonical_smiles'] == id_rd_to_oe['canonical_smiles'] == 'C(=CCl)F'
    assert id_oe_to_oe['canonical_isomeric_smiles'] == id_rd_to_oe['canonical_isomeric_smiles'] == 'C(=C\\Cl)\\F'
    assert id_oe_to_oe['canonical_explicit_hydrogen_smiles'] == id_rd_to_oe['canonical_explicit_hydrogen_smiles'] == '[H]C(=C([H])Cl)F'
    assert id_oe_to_oe['canonical_isomeric_explicit_hydrogen_smiles'] == id_rd_to_oe['canonical_isomeric_explicit_hydrogen_smiles'] == '[H]/C(=C(\\[H])/Cl)/F'
    assert id_oe_to_oe['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[H:5]/[C:1](=[C:2](\\[H:6])/[Cl:4])/[F:3]'
    assert id_rd_to_oe['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[H:1]/[C:5](=[C:6](\\[H:2])/[Cl:4])/[F:3]'

    assert id_oe_to_rd['canonical_smiles'] == id_rd_to_rd['canonical_smiles'] == 'FC=CCl'
    assert id_oe_to_rd['canonical_isomeric_smiles'] == id_rd_to_rd['canonical_isomeric_smiles'] == 'F/C=C\\Cl'
    assert id_oe_to_rd['canonical_explicit_hydrogen_smiles'] == id_rd_to_rd['canonical_explicit_hydrogen_smiles'] == '[H][C]([F])=[C]([H])[Cl]'
    assert id_oe_to_rd['canonical_isomeric_explicit_hydrogen_smiles'] == id_rd_to_rd['canonical_isomeric_explicit_hydrogen_smiles'] == '[H]/[C]([F])=[C](\\[H])[Cl]'
    assert id_oe_to_rd['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[C:1](=[C:2](\\[Cl:4])[H:6])(\\[F:3])[H:5]'
    assert id_rd_to_rd['canonical_isomeric_explicit_hydrogen_mapped_smiles'] == '[H:1]/[C:5]([F:3])=[C:6](\\[H:2])[Cl:4]'


@using_openeye
def _chemical_formula_oe(smiles):
    from openeye import oechem
    molecule = cmiles.utils.load_molecule(smiles, toolkit='openeye')
    oechem.OEAddExplicitHydrogens(molecule)
    return (cmiles.utils.mol_to_hill_molecular_formula(molecule))

@using_rdkit
def _chemical_formula_rd(smiles):
    from rdkit import Chem
    molecule = cmiles.utils.load_molecule(smiles, toolkit='rdkit')
    molecule = Chem.AddHs(molecule)
    return (cmiles.utils.mol_to_hill_molecular_formula(molecule))

@using_rdkit
@using_openeye
@pytest.mark.parametrize("backend", [_chemical_formula_oe, _chemical_formula_rd])
@pytest.mark.parametrize("smiles, bench", [("CCCC", "C4H10"), ("C", "CH4"), ("CC(Br)(Br)", "C2H4Br2"),
                                           ("[Li+].[O]=O", "LiO2")])
def test_molecule_formula(backend, smiles, bench):
    assert backend(smiles) == bench


def _test_unique_protomer(state1, state2):
    """Test unique protomer"""
    mol_1 = cmiles.utils.load_molecule(state1)
    mol_2 = cmiles.utils.load_molecule(state2)
    assert cmiles.get_unique_protomer(mol_1) == cmiles.get_unique_protomer(mol_2)


def _test_unique_tautomer(state1, state2):
    """Test standardize tautomer rdkit"""
    assert cmiles.standardize_tautomer(state1) == cmiles.standardize_tautomer(state2)

@using_rdkit
@using_openeye
@pytest.mark.parametrize('state1, state2',
                          [
                           ('CN4CCN(c2nc1cc(Cl)ccc1[nH]c3ccccc23)CC4', 'CN4CCN(c2[nH]c1cc(Cl)ccc1nc3ccccc23)CC4'), # 1,5 tautomerism
                           ('CC=CC(=O)C', 'CC(=CC=C)O'), #keto-enol
                           ('NC=O', 'N=CO'), # amide - imidic
                           ('C1CC(=O)NC1', 'C1CC(O)=NC1'), # lactam-lactim

                           ('C1=CN=CNC1=O', 'C1=CNC=NC1=O'), # pyrimodone
                           ('C1=CN=CNC1=O', 'C1=CN=CN=C1O'),# pyrimodol - pytimodone
                           ('C1=NC=NN1','N1C=NN=C1' ), #triazole
                           ('C1(=NC(=NC(=N1)O)O)O', 'C1(NC(NC(N1)=O)=O)=O'), # cyanic - cyanuric
                           ('CC(=O)S[H]', 'CC(=S)O[H]'),
                           ('C[N](=O)=O', 'C[N+]([O-])=O'),# nitromethane
                           ('CN=[N]#N', 'CN=[N+]=[N-]'),  # azide
                           ('C[S](C)(=O)=O', 'C[S++](C)([O-])[O-]'), # sulfone
                           ('C[P](C)(=O)[O-]', 'C[P+](C)([O-])[O-]'), # phosphinate
                           ('OC=C', 'O=CC'), # keto-enol
                           ('[H]NC=C', '[H]N=CC'), # imine - enamine
                           ('CN=O', 'C=NO'), # nitoroso-oxime
                           ('C1=CC(NN1)=O','C1=NNC(C1)=O' ),# pyrazolone
                           ('C1=CC(NN1)=O', 'C1(=CC=N[N]1[H])O'), # pyrazolone
                           ('C1(=NC=CNC=C1)C', 'C1(=CC=NC=CN1)C'),# diazepine
                           ('C1=CC=C(O1)O','C1=CCC(O1)=O' ), #furanol-furanone
                           ('C1=CC(=CO1)O', 'C1=CC(CO1)=O')#  furanol - furanone
                           ])
@pytest.mark.parametrize('backend', [_test_unique_protomer, _test_unique_tautomer])
def test_universal_id(backend, state1, state2):
    backend(state1, state2)


# only captured with openeye - unique protomer also included ionization states. rdkit only standardizes for tautomers.
@using_openeye
@pytest.mark.parametrize("state1, state2",
                         [('CC(=O)O','CC(=O)[O-]'), # protonation states
                          ('NCC(=O)O', '[NH3+]CC(=O)[O-]'),# amino acid
                          ('CC(=O)S[H]', 'CC(=O)[S-]'),
                          ('CC(=O)S[H]', 'CC(=S)[O-]'),
                         ])
def test_unique_protomers(state1, state2):
    _test_unique_protomer(state1, state2)

# only captured with redkit. captures indoles. Neither openeye nor rdkit capture isoindoles
@using_rdkit
@pytest.mark.parametrize("state1, state2",
                         [('CP(=O)(C)S', 'CP(=S)(C)O'), # mesomer
                          ('CC(=[NH2+])N','C[C+](N)N'), # mesomer
                          ('C=C=O', 'C#CO'), # keten-ynol
                          ('c1ccc2c(c1)cc([nH]2)O', 'c1ccc2c(c1)CC(=O)N2'), # indole_ol - indole-one
                          ('c1ccc2c(c1)cc([nH]2)N', 'c1ccc2c(c1)CC(=N2)N')# indole - indole
                         ])
def test_unique_tautomers(state1, state2):
    _test_unique_tautomer(state1, state2)


@pytest.mark.parametrize('toolkit', toolkits_name)
def test_permute_xyz(toolkit):

    hooh = {
        'symbols': ['H', 'O', 'O', 'H'],
        'geometry': [
             1.84719633,  1.47046223,  0.80987166,
             1.3126021,  -0.13023157, -0.0513322,
            -1.31320906,  0.13130216, -0.05020593,
            -1.83756335, -1.48745318,  0.80161212
        ],
        'name': 'HOOH',
        'connectivity': [[0, 1, 1], [1, 2, 1], [2, 3, 1]],
        'molecular_multiplicity': 1
    }

    mol_ids, permuted_hooh = cmiles.get_molecule_ids(hooh, toolkit, permute_xyz=True)

    assert hooh['geometry'] != permuted_hooh['geometry']

    mol = cmiles.utils.mol_from_json(hooh, toolkit=toolkit)
    atom_map = cmiles.utils.get_atom_map(molecule=mol, mapped_smiles=mol_ids['canonical_isomeric_explicit_hydrogen_mapped_smiles'])

    json_geom = np.asarray(hooh['geometry']).reshape(int(len(hooh['geometry'])/3), 3)
    permuted_geom = np.asarray(permuted_hooh['geometry']).reshape(int(len(hooh['geometry'])/3), 3)
    for m in atom_map:
        for i in range(3):
            assert json_geom[atom_map[m]][i] == pytest.approx(permuted_geom[m-1][i], 0.0000001)

@pytest.mark.parametrize('toolkit', toolkits_name)
def test_provenance(toolkit):
    provenance = cmiles.get_molecule_ids('[H]C([H])([H])C([H])([H])C([H])([H])C([H])([H])[H]', toolkit=toolkit)['provenance']
    assert provenance.split('_')[2] == toolkit



"""

"""
import rdkit as toolkit
from rdkit import Chem
from .utils import _symbols, ANGSROM_2_BOHR
import warnings
import copy


def mol_from_json(symbols, connectivity, geometry, permute_xyz=False):
    """
    Generate RDkit.Chem.Mol from QCSchema molecule specs.
    Parameters
    ----------
    inp_molecule: dict
        Must include symbols and connectivity. Geometry is optional. If geometry is given, stereochemistry will be taken
        from coordinates

    Returns
    -------
    molecule: rdkit.Chem.Mol
    """

    from rdkit.Geometry.rdGeometry import Point3D

    _BO_DISPATCH_TABLE = {1: Chem.BondType.SINGLE, 2: Chem.BondType.DOUBLE, 3: Chem.BondType.TRIPLE}

    geometry = geometry.reshape(int(len(geometry)/3), 3)
    conformer = Chem.Conformer(len(symbols))
    has_geometry = True

    molecule = Chem.Mol()
    em = Chem.RWMol(molecule)
    for i, s in enumerate(symbols):
        atom = em.AddAtom(Chem.Atom(_symbols[s]))
        atom_position = Point3D(geometry[i][0], geometry[i][1], geometry[i][2])
        conformer.SetAtomPosition(atom, atom_position)

    # Add connectivity
    for bond in connectivity:
        bond_type = _BO_DISPATCH_TABLE[bond[-1]]
        em.AddBond(bond[0], bond[1], bond_type)

    molecule = em.GetMol()
    try:
        Chem.SanitizeMol(molecule)
    except:
        raise RuntimeError("Could not sanitize molecule")

    # Add coordinates
    if has_geometry:
        initial_conformer_id = molecule.AddConformer(conformer, assignId=True)
        # Assign stereochemistry from coordinates
        from rdkit.Chem import rdmolops
        rdmolops.AssignStereochemistryFrom3D(molecule, confId=initial_conformer_id, replaceExistingTags=True)
        if not permute_xyz:
            # Add a tag to keep current order
            molecule.SetProp("_json_geometry", '1')

    return molecule


def canonical_order_atoms(molecule, h_last=True):
    """
    Canonical order atoms in RDKit molecule. Eaach atom in the molecule is given a map index that corresponds to the RDkit
    rank for that atom (+1). RDKit atom ranking ranks hydrogens first and then the heavy atoms. When
    h_last is set to True, the map indices are reordered to put hydrogens after the heavy atoms.
    Parameters
    ----------
    molecule: rdkit mol
    h_last: bool, optional, default is True

    Returns
    -------
    molecule: rdkit molecule with map indices that correspond to the atom canonical rank
    """

    # Check if molecule already has map. If it does, remove map because Chem.CanonicalRankAtoms uses map indices in
    # ranking
    if has_atom_map(molecule):
        remove_atom_map(molecule)

    # Add explicit hydrogen
    molecule = Chem.AddHs(molecule)
    heavy_atoms = 0
    hydrogens = 0
    ranks = list(Chem.CanonicalRankAtoms(molecule, breakTies=True))
    for i, j in enumerate(ranks):
        atom = molecule.GetAtomWithIdx(i)
        atom.SetAtomMapNum(j+1)
        if atom.GetAtomicNum() != 1:
            # heavy atom
            heavy_atoms +=1
        else:
            # hydrogen
            hydrogens +=1

    if h_last:
        # reorder map to put hydrogen last
        for atom in molecule.GetAtoms():
            map_idx = atom.GetAtomMapNum()
            if atom.GetAtomicNum() !=1:
                atom.SetAtomMapNum(map_idx - hydrogens)
            else:
                atom.SetAtomMapNum(map_idx + heavy_atoms)
    return molecule


def mol_to_smiles(molecule, isomeric=True, explicit_hydrogen=True, mapped=True):
    """
    Generate canonical SMILES with RDKit

    Parameters
    ----------
    molecule: RDKit Chem.rdchem.Mol instance
        The molecule to generate SMILES for
    isomeric: bool
        If True, SMILES will have isomeric information. If molecule already has isomeric information, this will be retained.
        If no isomeric information exists, this function will perceive it and assign the CW (clockwise) flag for chiral
        centers and the E-isomer for stereo bonds.
    explicit_hydrogen: bool
        If True, SMILES will have explicit hydrogens
    mapped: bool
        If True, SMILES will have map indices. (+1 because the map is 1 indexed)

    Returns
    -------
    smiles: str
        The canonical SMILES

    """
    if mapped and not explicit_hydrogen:
        raise Warning("Tagged SMILES must include hydrogens to retain order")

    if mapped and not isomeric:
        raise Warning("Tagged SMILES must include stereochemistry ")

    if explicit_hydrogen:
        # Add explicit hydrogens
        molecule = Chem.AddHs(molecule)
    if not explicit_hydrogen:
        molecule = Chem.RemoveHs(molecule)

    try:
        json_geometry = int(molecule.GetProp("_json_geometry"))
    except KeyError:
        json_geometry = False

    if isomeric and not has_stereo_defined(molecule):
        raise ValueError("Some stereochemistry is not defined")

    # Get canonical order for map
    if mapped:
        if json_geometry:
            # keep original ordering
            #ToDo this looks like a potential bug that only json_geometry gets atom maps
            for i in range(molecule.GetNumAtoms()):
                molecule.GetAtomWithIdx(i).SetAtomMapNum(i+1)
        else:
            molecule = canonical_order_atoms(molecule)

    smiles = Chem.MolToSmiles(molecule, allHsExplicit=explicit_hydrogen, isomericSmiles=isomeric, canonical=True)
    return smiles

def generate_index(mol, labeled_atoms):
    """
    Generate SMILES with labeled atoms

    Parameters
    ----------
    mol : rdmol
        Must have atom indices that correspond to labeled atoms
    labeled_atoms : tuple of ints
        ints should be -1 of atom map

    Returns
    -------
    labeled SMILES

    """
    # Remove explicit H
    mol = Chem.RemoveHs(mol)
    for a in mol.GetAtoms():
        if not a.GetAtomMapNum()-1 in labeled_atoms:
            a.SetAtomMapNum(0)

    for a in mol.GetAtoms():
        m = a.GetAtomMapNum() -1
        if m in labeled_atoms:
            i = labeled_atoms.index(m)
            a.SetAtomMapNum(i+1)
    return Chem.MolToSmiles(mol)

def get_connectivity_table(molecule, inverse_map):
    """

    Parameters
    ----------
    molecule
    inverse_map

    Returns
    -------

    """
    # Remove aromaticity
    Chem.Kekulize(molecule)
    connectivity_table = [[inverse_map[bond.GetBeginAtomIdx()]-1, inverse_map[bond.GetEndAtomIdx()]-1, bond.GetBondTypeAsDouble()]
                              for bond in molecule.GetBonds()]
    return connectivity_table



def get_atom_map(molecule, mapped_smiles):
    """

    Parameters
    ----------
    molecule
    mapped_smiles

    Returns
    -------

    """
    # Check for mapping in mapped_smiles
    mapped_mol = Chem.MolFromSmiles(mapped_smiles)
    if not has_atom_map(mapped_mol):
        raise ValueError("mapped SMILES must have map for every heavy atom and hydrogen")
    # Check if molecule has explicit H
    if not has_explicit_hydrogen(molecule):
        molecule = Chem.AddHs(molecule)

    mapped_pattern = Chem.MolFromSmarts(mapped_smiles)
    if molecule.HasSubstructMatch(mapped_pattern):
        idx_pattern_order = molecule.GetSubstructMatch(mapped_pattern)
    else:
        raise RuntimeError("Substrucure match failed for {}, SMARTS: {}".format(Chem.MolToSmiles(molecule), mapped_smiles))
    atom_map = {}
    for i, j in enumerate(idx_pattern_order):
        atom_map[mapped_pattern.GetAtomWithIdx(i).GetAtomMapNum()] = j

    return atom_map


def get_map_ordered_geometry(molecule, atom_map):
    """

    Parameters
    ----------
    molecule

    Returns
    -------

    """
    try:
        if not molecule.GetConformer().Is3D():
            raise ValueError("Molecule must have 3D coordinates for generating QCSchema molecule")
    except ValueError:
        raise ValueError("Molecule must have 3D coordinates for generating QCSchema molecule")

    symbols = []
    geometry = []
    for mapping in range(1, molecule.GetNumAtoms()+1):
        idx = atom_map[mapping]
        atom = molecule.GetAtomWithIdx(idx)
        syb = atom.GetSymbol()
        symbols.append(syb)
        pos = molecule.GetConformer().GetAtomPosition(atom.GetIdx())
        geometry.extend(getattr(pos, c) * ANGSROM_2_BOHR for c in ['x', 'y', 'z'])

    return symbols, geometry


def has_explicit_hydrogen(molecule):
    """

    Parameters
    ----------
    molecule

    Returns
    -------

    """
    explicit = True
    for a in molecule.GetAtoms():
        if a.GetImplicitValence() > 0 or a.GetNumExplicitHs() > 0:
            explicit = False

    return explicit


def add_explicit_hydrogen(molecule):
    molecule = Chem.AddHs(molecule)
    return molecule


def get_symbols(molecule):
    return [a.GetSymbol() for a in molecule.GetAtoms()]


def has_stereo_defined(molecule):
    #ToDo Fix this function for carbons with explicit hydrogen and imine with implicit hydrogens
    """

    Parameters
    ----------
    molecule

    Returns
    -------

    """
    unspec_chiral = False
    unspec_db = False
    problematic_atoms = list()
    problematic_bonds = list()
    # remove map indices and store. Then create molecule with SMILES without map indices and check for stereo on that.
    # This step is needed because even when map indices are stored in the data of the atoms, it is used for assigning potential
    # stereo centers. If a molecule does not have map indices but has data on the atoms, it will get flagged
    had_atom_map = False
    if has_atom_map(molecule):
        had_atom_map = True
        remove_atom_map(molecule)
    a = Chem.rdmolfiles.SmilesParserParams()
    a.removeHs = False
    mol_copy = Chem.MolFromSmiles(Chem.MolToSmiles(molecule), a)
    # restore map indices
    if had_atom_map:
        restore_atom_map(molecule)
    chiral_centers = Chem.FindMolChiralCenters(mol_copy, includeUnassigned=True)
    for center in chiral_centers:
        atom_id = center[0]
        if center[-1] == '?':
            unspec_chiral = True
            problematic_atoms.append((atom_id, mol_copy.GetAtomWithIdx(atom_id).GetSmarts()))

    # Find potential stereo bonds that are unspecified
    Chem.FindPotentialStereoBonds(mol_copy)
    for bond in mol_copy.GetBonds():
        if bond.GetStereo() == Chem.BondStereo.STEREOANY:
            unspec_db = True
            problematic_bonds.append((bond.GetBeginAtom().GetSmarts(), bond.GetSmarts(),
                                                bond.GetEndAtom().GetSmarts()))
    if unspec_chiral or unspec_db:
        warnings.warn("Stereochemistry is unspecified. Problematic atoms {}, problematic bonds {}".format(
                problematic_atoms, problematic_bonds))
        return False
    else:
        return True


def has_atom_map(molecule):
    """
    Checks if any atom has map indices. Will return True if even only one atom has a map index
    Parameters
    ----------
    molecule

    Returns
    -------

    """
    IS_MAPPED = False
    for atom in molecule.GetAtoms():
            if atom.GetAtomMapNum() != 0:
                IS_MAPPED = True
                return IS_MAPPED
    return IS_MAPPED


def is_missing_atom_map(molecule):
    """
    Checks if any atom in molecule is missing a map index. If even only one atom is missing a map index will return True

    Parameters
    ----------
    molecule

    Returns
    -------

    """
    MISSING_ATOM_MAP = False
    for atom in molecule.GetAtoms():
            if atom.GetAtomMapNum() == 0:
                MISSING_ATOM_MAP = True
                return MISSING_ATOM_MAP
    return MISSING_ATOM_MAP


def remove_atom_map(molecule, keep_map_data=True):
    """
    Remove atom map but store it in atom data.
    Parameters
    ----------
    molecule

    Returns
    -------

    """
    for atom in molecule.GetAtoms():
        if atom.GetAtomMapNum() != 0:
            if keep_map_data:
                atom.SetProp('_map_idx', str(atom.GetAtomMapNum()))
            atom.SetAtomMapNum(0)


def restore_atom_map(molecule):
    """
    Restore atom map from atom data
    Parameters
    ----------
    molecule: OEMol
        Must have 'MapIdx' in atom data dictionary
    """
    for atom in molecule.GetAtoms():
        if atom.HasProp('_map_idx'):
            atom.SetAtomMapNum(int(atom.GetProp('_map_idx')))

def add_atom_map(molecule, in_place=False):
    """
    Add canonical ordered atom map to existing molecule

    Parameters
    ----------
    molecule : rdkit.Chem.Mol
    in_place : bool, default False

    Returns
    -------
    rdkit.Chem.Mol with atom map
    """
    # First canonical order atoms
    molecule =  canonical_order_atoms(molecule)

    for i in range(molecule.GetNumAtoms()):
        molecule.GetAtomWithIdx(i).SetAtomMapNum(i+1)
    return molecule


def is_map_canonical(molecule):
    """
    Check if map indices on molecule is in canonical order.
    Note:
    This only checks map indices on heavy atoms.

    Parameters
    ----------
    molecule

    Returns
    -------

    """
    molcopy = copy.deepcopy(molecule)
    # reorder molcopy
    canonical_ordered_mol = canonical_order_atoms(molcopy)

    # check that both mapped SMILES are equal
    smiles_1 = Chem.MolToSmiles(molecule, canonical=True)
    smiles_2 = Chem.MolToSmiles(canonical_ordered_mol, canonical=True)
    return smiles_1 == smiles_2



"""Test util functions"""

from cmiles import utils
import pytest
import numpy as np

toolkits = list()
toolkits_name = list()
if utils.has_rdkit:
    from cmiles import _cmiles_rd
    from rdkit import Chem
    toolkits.append(_cmiles_rd)
    toolkits_name.append('rdkit')
if utils.has_openeye:
    from cmiles import _cmiles_oe
    from openeye import oechem
    toolkits.append(_cmiles_oe)
    toolkits_name.append('openeye')

using_rdkit = pytest.mark.skipif(not utils.has_rdkit, reason="Cannot run without RDKit")
using_openeye = pytest.mark.skipif(not utils.has_openeye, reason="Cannot run without OpenEye")


@pytest.mark.parametrize('toolkit', toolkits_name)
def test_load_molecule(toolkit):
    """Test load molecules"""
    mol = utils.load_molecule('[H]C([H])([H])C([H])([H])C([H])([H])C([H])([H])[H]', toolkit=toolkit)
    if toolkit == 'openeye':
        from openeye import oechem
        assert oechem.OEMolToSmiles(mol) == 'CCCC'
    if toolkit == 'rdkit':
        from rdkit import Chem
        assert Chem.MolToSmiles(mol) == '[H]C([H])([H])C([H])([H])C([H])([H])C([H])([H])[H]'


@pytest.mark.parametrize('toolkit', toolkits_name)
@pytest.mark.parametrize('input, output', [('[H:3][C:1]([H:4])([H:5])[C:2]([H:6])([H:7])[H:8]', True),
                                           ('[H:3][C:1]([H])([H:5])[C]([H])([H:7])[H:8]', True),
                                           ('CCCC', False)])
def test_is_mapped(toolkit, input, output):
    """Test is mapped"""
    mapped_mol = utils.load_molecule(input, toolkit=toolkit)
    assert utils.has_atom_map(mapped_mol) == output
    utils.remove_atom_map(mapped_mol)
    assert utils.has_atom_map(mapped_mol) == False


@pytest.mark.parametrize('toolkit', toolkits_name)
@pytest.mark.parametrize('input, output', [('[H:3][C:1]([H:4])([H:5])[C:2]([H:6])([H:7])[H:8]', False),
                                           ('[H:3][C:1]([H:4])([H:5])[C]([H:6])([H:7])[H:8]', True),
                                           ('CCCC', True)])
def test_is_missing_map(toolkit, input, output):
    #ToDo - Known problem that RDKit does not add explicit H to molecules even with explicit H SMILES so if map of H is missing it will not pick it up
    mol = utils.load_molecule(input, toolkit=toolkit)
    assert utils.is_missing_atom_map(mol) == output


@pytest.mark.parametrize('toolkit_str', toolkits_name)
def test_mol_from_json(toolkit_str):
    """Test oemol from json"""
    import numpy as np
    hooh = {
        'symbols': ['H', 'O', 'O', 'H'],
        'geometry': [
             1.84719633,  1.47046223,  0.80987166,
             1.3126021,  -0.13023157, -0.0513322,
            -1.31320906,  0.13130216, -0.05020593,
            -1.83756335, -1.48745318,  0.80161212
        ],
        'name': 'HOOH',
        'connectivity': [[0, 1, 1], [1, 2, 1], [2, 3, 1]],
    }
    mol = utils.load_molecule(hooh, toolkit=toolkit_str)
    if toolkit_str == 'openeye':
        assert mol.GetMaxAtomIdx() == 4
        assert mol.GetMaxBondIdx() == 3
        coordinates = mol.GetCoords()
    if toolkit_str == 'rdkit':
        assert mol.GetNumAtoms() == 4
        assert mol.GetNumBonds() == 3
        coordinates = mol.GetConformer().GetPositions()
    geometry = np.array(hooh['geometry'], dtype=float).reshape(int(len(hooh['geometry'])/3), 3)*utils.BOHR_2_ANGSTROM
    for i in range(len(coordinates)):
        for j in range(3):
            assert coordinates[i][j] == pytest.approx(geometry[i][j], 0.0000001)


@pytest.mark.parametrize('toolkit_name', toolkits_name)
@pytest.mark.parametrize("input1, input2", [('C[C@](N)(O)F', 'CC(N)(O)F'),
                                   ('C(=C/Cl)\\F', 'C(=CCl)F'),
                                   ('CN(C)C/C=C/C(=O)NC1=C(C=C2C(=C1)C(=NC=N2)NC3=CC(=C(C=C3)F)Cl)O[C@H]4CCOC4',
                                   'CN(C)CC=CC(=O)NC1=C(C=C2C(=C1)C(=NC=N2)NC3=CC(=C(C=C3)F)Cl)O[C@H]4CCOC4'),
                                   ('CN(C)C/C=C/C(=O)NC1=C(C=C2C(=C1)C(=NC=N2)NC3=CC(=C(C=C3)F)Cl)O[C@H]4CCOC4',
                                   'CN(C)C/C=C/C(=O)NC1=C(C=C2C(=C1)C(=NC=N2)NC3=CC(=C(C=C3)F)Cl)OC4CCOC4')])
def test_has_stereochemistry(input1, input2, toolkit_name):
    mol = utils.load_molecule(input1, toolkit_name)
    if toolkit_name == 'openeye':
        from openeye import oechem
        oechem.OEAddExplicitHydrogens(mol)
    if toolkit_name == 'rdkit':
        from rdkit import Chem
        mol = Chem.AddHs(mol)
    assert utils.has_stereo_defined(mol) == True

    mol = utils.load_molecule(input2, toolkit_name)
    if toolkit_name == 'openeye':
        from openeye import oechem
        oechem.OEAddExplicitHydrogens(mol)
    if toolkit_name == 'rdkit':
        from rdkit import Chem
        mol = Chem.AddHs(mol)
    with pytest.warns(UserWarning):
        utils.has_stereo_defined(mol)


@using_openeye
def test_canonical_order_oe():
    """Test canonical atom order"""
    mol = oechem.OEMol()
    oechem.OESmilesToMol(mol, 'HOOH')

    # add map
    for atom in mol.GetAtoms():
        atom.SetMapIdx(atom.GetIdx() + 1)

    assert oechem.OEMolToSmiles(mol) == '[H:1][O:2][O:3][H:4]'

    mol_2 = _cmiles_oe.canonical_order_atoms(mol, in_place=False)
    for atom in mol.GetAtoms():
        atom.SetMapIdx(atom.GetIdx() + 1)
    for atom in mol_2.GetAtoms():
        atom.SetMapIdx(atom.GetIdx() + 1)
    assert oechem.OEMolToSmiles(mol) == '[H:1][O:2][O:3][H:4]'
    assert oechem.OEMolToSmiles(mol_2) == '[H:3][O:1][O:2][H:4]'

    _cmiles_oe.canonical_order_atoms(mol, in_place=True)
    for atom in mol.GetAtoms():
        atom.SetMapIdx(atom.GetIdx() + 1)
    assert oechem.OEMolToSmiles(mol) == '[H:3][O:1][O:2][H:4]'


@using_rdkit
def test_canonical_order_rd():
    """Test canonical atom order"""
    mol = Chem.MolFromSmiles('CO')

    # add map
    mol = Chem.AddHs(mol)
    for i in range(mol.GetNumAtoms()):
        mol.GetAtomWithIdx(i).SetAtomMapNum(i+1)

    assert Chem.MolToSmiles(mol) == '[C:1]([O:2][H:6])([H:3])([H:4])[H:5]'

    mol_2 = _cmiles_rd.canonical_order_atoms(mol, h_last=False)
    assert Chem.MolToSmiles(mol_2) == '[H:1][O:5][C:6]([H:2])([H:3])[H:4]'

    mol_3 = _cmiles_rd.canonical_order_atoms(mol)
    assert Chem.MolToSmiles(mol_3) == '[O:1]([C:2]([H:4])([H:5])[H:6])[H:3]'


@pytest.mark.parametrize('toolkit_name', toolkits_name)
@pytest.mark.parametrize('input, output', [('COC(C)c1c(Cl)ccc(F)c1Cl', False),
                                           ('C[C@@H](c1c(ccc(c1Cl)F)Cl)OC', False),
                                           ('O=O', True),
                                           ('[CH3:1][CH2:3][CH2:4][CH3:2]', False),
                                           ('[H]c1c(c(c(c(c1F)Cl)[C@]([H])(C([H])([H])[H])OC([H])([H])[H])Cl)[H]', True),
                                           ('[C:1]([O:2][H:6])([H:3])([H:4])[H:5]', True)])
def test_explicit_h(input, output, toolkit_name):
    """Test input SMILES for explicit H"""
    mol = utils.load_molecule(input, toolkit=toolkit_name)
    assert utils.has_explicit_hydrogen(mol) == output


@pytest.mark.parametrize('toolkit', toolkits_name)
@pytest.mark.parametrize("smiles", ['CN(C)C(=N)NC(=N)N', 'N=CO'])
def test_chiral_bond_exception(smiles, toolkit):
    """ Test bonds to ignore """
    mol = utils.load_molecule(smiles, toolkit)
    if toolkit == 'openeye':
        from openeye import oechem
        oechem.OEAddExplicitHydrogens(mol)
    if toolkit == 'rdkit':
        from rdkit import Chem
        mol = Chem.AddHs(mol)
    with pytest.warns(UserWarning):
        utils.has_stereo_defined(mol)


@pytest.mark.parametrize('toolkit, toolkit_name', list(zip(toolkits, toolkits_name)))
def test_get_atom_map(toolkit, toolkit_name):
    smiles = 'C[C@@H](c1c(ccc(c1Cl)F)Cl)OC'
    mol = utils.load_molecule(smiles, toolkit_name)
    if toolkit_name == 'openeye':
        from openeye import oechem
        oechem.OEAddExplicitHydrogens(mol)
        for a in mol.GetAtoms():
            a.SetMapIdx(a.GetIdx()+1)
        mapped_smiles = oechem.OEMolToSmiles(mol)

    if toolkit_name == 'rdkit':
        from rdkit import Chem
        mol = Chem.AddHs(mol)
        for a in mol.GetAtoms():
            a.SetAtomMapNum(a.GetIdx()+1)
        mapped_smiles = Chem.MolToSmiles(mol)


    atom_map = utils.get_atom_map(mol, mapped_smiles)

    for m in atom_map:
        assert m == (atom_map[m] + 1)

@using_openeye
@pytest.mark.parametrize('smiles', ['OCCO', 'C(CO)O', '[H]C([H])(C([H])([H])O[H])O[H]',
                                    '[H:5][C:1]([H:6])([C:2]([H:7])([H:8])[O:4][H:10])[O:3][H:9]',
                                   '[H][O][C]([H])([H])[C]([H])([H])[O][H]',
                                    '[O:1]([C:3]([C:4]([O:2][H:6])([H:9])[H:10])([H:7])[H:8])[H:5]'])
def test_atom_map(smiles):
    """Test that atom map orders geometry the same way every time no matter the SMILES used to create the molecule"""
    import cmiles
    mapped_smiles = '[H:5][C:1]([H:6])([C:2]([H:7])([H:8])[O:4][H:10])[O:3][H:9]'
    mol_id_oe = cmiles.get_molecule_ids(mapped_smiles, toolkit='openeye')
    oemol = utils.load_molecule(mapped_smiles, toolkit='openeye')
    mapped_symbols = ['C', 'C', 'O', 'O', 'H', 'H', 'H', 'H', 'H', 'H']
    mapped_geometry = [-1.6887193912042044, 0.8515190939276903, 0.8344587822904272, -4.05544806361675, -0.3658269566455062,
                       -0.22848169646448416, -1.6111611950422127, 0.4463128276938808, 3.490617694146934, -3.97756355964586,
                       -3.0080934853087373, 0.25948499322223956, -1.6821252026076652, 2.891135395246369, 0.4936556190978574,
                       0.0, 0.0, 0.0, -4.180315034973438, -0.09210893239246959, -2.2748227320305525, -5.740516456782416,
                       0.4115539217904015, 0.6823267491485907, -0.07872657410528058, 1.2476492272884379, 4.101615944163073,
                       -5.514569080545831, -3.7195945404657222, -0.4441653010509862]

    mol = cmiles.utils.load_molecule(smiles, toolkit='openeye')
    if not utils.has_explicit_hydrogen(mol):
        mol = utils.add_explicit_hydrogen(mol)
    atom_map = utils.get_atom_map(mol, mapped_smiles=mapped_smiles)
    # use the atom map to add coordinates to molecule. First reorder mapped geometry to order in molecule
    mapped_coords = np.array(mapped_geometry, dtype=float).reshape(int(len(mapped_geometry)/3), 3)
    coords = np.zeros((mapped_coords.shape))
    for m in atom_map:
        coords[atom_map[m]] = mapped_coords[m-1]
    # flatten
    coords = coords.flatten()
    # convert to Angstroms
    coords = coords*utils.BOHR_2_ANGSTROM
    # set coordinates in oemol
    mol.SetCoords(coords)
    mol.SetDimension(3)

    # Get new atom map
    atom_map = utils.get_atom_map(mol, mapped_smiles)
    symbols, geometry = _cmiles_oe.get_map_ordered_geometry(mol, atom_map)
    assert geometry == mapped_geometry
    assert symbols == mapped_symbols


@pytest.mark.parametrize('mapped_smiles, expected_table', [('[H:5][C:1]([H:6])([H:7])[C:3]([H:11])([H:12])[C:4]([H:13])([H:14])[C:2]([H:8])([H:9])[H:10]',
                                                           np.array([[0, 2, 1], [1, 3, 1],[2, 3, 1],[0, 4, 1],[0, 5, 1],
                                                                    [0, 6, 1],[1, 7, 1],[1, 8, 1],[1, 9, 1],[2, 10, 1],
                                                                    [2, 11, 1],[3, 12, 1],[3, 13, 1]])),
                                                           ('[H:7][c:1]1[c:2]([c:4]([o:5][c:3]1[H:9])[O:6][H:10])[H:8]',
                                                            np.array([[2, 4, 1], [2, 0, 2], [0, 1, 1], [1, 3, 2],
                                                                      [3, 4, 1], [3, 5, 1], [2, 8, 1], [0, 6, 1],
                                                                      [1, 7, 1], [5, 9, 1]]))])
@pytest.mark.parametrize('toolkit', toolkits_name)
def test_connectivity(mapped_smiles, expected_table, toolkit):
    """Test connectivity table"""
    molecule = utils.load_molecule(mapped_smiles, toolkit)

    atom_map = utils.get_atom_map(molecule, mapped_smiles)
    connectivity_table = utils.get_connectivity_table(molecule, atom_map)

    for bond in connectivity_table:
        xi = np.isin(expected_table, bond[:2])
        match = np.where(np.array([i[:2].sum() for i in xi]) == 2)[0]
        # assert that a match was found and only one was found
        assert len(match) == 1
        # assert that bond order is the same
        assert expected_table[match][0][-1] == bond[-1]


@pytest.mark.parametrize('toolkit, toolkit_name', list(zip(toolkits, toolkits_name)))
@pytest.mark.parametrize('permute', [True, False])
def test_map_order_geometry(permute, toolkit, toolkit_name):
    """Test map ordered geometry"""
    hooh = {
        'symbols': ['H', 'O', 'O', 'H'],
        'geometry': [
             1.84719633,  1.47046223,  0.80987166,
             1.3126021,  -0.13023157, -0.0513322,
            -1.31320906,  0.13130216, -0.05020593,
            -1.83756335, -1.48745318,  0.80161212
        ],
        'name': 'HOOH',
        'connectivity': [[0, 1, 1], [1, 2, 1], [2, 3, 1]],
    }
    mol = utils.load_molecule(hooh, toolkit=toolkit_name, permute_xyz=permute)
    mapped_smiles = utils.mol_to_smiles(mol, isomeric=True, explicit_hydrogen=True, mapped=True)
    atom_map = utils.get_atom_map(mol, mapped_smiles)
    symbols, geometry = toolkit.get_map_ordered_geometry(mol, atom_map)

    json_geom = np.asarray(hooh['geometry']).reshape(int(len(geometry)/3), 3)
    geometry_array = np.asarray(geometry).reshape(int(len(geometry)/3), 3)

    for m in atom_map:
        for i in range(3):
            assert json_geom[atom_map[m]][i] == pytest.approx(geometry_array[m-1][i], 0.0000001)
    if not permute:
        assert hooh['geometry'] == pytest.approx(geometry, 0.0000001)


@pytest.mark.parametrize('toolkit', toolkits_name)
def test_permute_json(toolkit):
    """Test permute json xyz, symbols and connectivity to match mapped smiles"""
    molecule_ids = {'canonical_smiles': 'OO',
                    'canonical_isomeric_smiles': 'OO',
                    'canonical_isomeric_explicit_hydrogen_smiles': '[H]OO[H]',
                    'canonical_explicit_hydrogen_smiles': '[H]OO[H]',
                    'canonical_isomeric_explicit_hydrogen_mapped_smiles': '[H:3][O:1][O:2][H:4]',
                    'unique_protomer_representation': 'OO',
                    'unique_tautomer_representation': 'OO',
                    'provenance': 'cmiles_v0.1.1+59.g0b7a12d.dirty_openeye_2018.Oct.b7',
                    'standard_inchi': 'InChI=1S/H2O2/c1-2/h1-2H',
                    'inchi_key': 'MHAJPDPJQMAIIY-UHFFFAOYSA-N',
                    'molecular_formula': 'H2O2'}
    hooh = {
        'symbols': ['H', 'O', 'O', 'H'],
        'geometry': [
             1.84719633,  1.47046223,  0.80987166,
             1.3126021,  -0.13023157, -0.0513322,
            -1.31320906,  0.13130216, -0.05020593,
            -1.83756335, -1.48745318,  0.80161212
        ],
        'name': 'HOOH',
        'connectivity': [[0, 1, 1], [1, 2, 1], [2, 3, 1]],
        'molecular_multiplicity': 1
    }
    mapped_smiles = molecule_ids['canonical_isomeric_explicit_hydrogen_mapped_smiles']
    mol = utils.mol_from_json(hooh, toolkit=toolkit)
    atom_map = utils.get_atom_map(mol, '[H:3][O:1][O:2][H:4]')
    permuted_hooh = utils.permute_qcschema(hooh, mapped_smiles, toolkit=toolkit)

    json_geom = np.asarray(hooh['geometry']).reshape(int(len(hooh['geometry'])/3), 3)
    permuted_geom = np.asarray(permuted_hooh['geometry']).reshape(int(len(hooh['geometry'])/3), 3)

    assert hooh['geometry'] != permuted_hooh['geometry']
    for m in atom_map:
        for i in range(3):
            assert json_geom[atom_map[m]][i] == pytest.approx(permuted_geom[m-1][i], 0.0000001)


@pytest.mark.parametrize('toolkit', toolkits_name)
def test_get_atom_map_mapped_smiles(toolkit):
    smiles_1 = '[H]C([H])(C([H])([H])O[H])O[H]'
    smiles_2 = '[H:5][C:1]([H:6])([C:2]([H:7])([H:8])[O:4][H:10])[O:3][H:9]'
    mol_1 = utils.load_molecule(smiles_1, toolkit=toolkit)
    if not utils.has_explicit_hydrogen(mol_1):
        mol_1 = utils.add_explicit_hydrogen(mol_1)
    mol_2 = utils.load_molecule(smiles_2, toolkit=toolkit)
    if not utils.has_explicit_hydrogen(mol_2):
        mol_2 = utils.add_explicit_hydrogen(mol_2)


@pytest.mark.parametrize('toolkit', toolkits_name)
def test_remove_restore_atom_map(toolkit):
    mapped_smiles = '[H:5][C:1]([H:6])([C:2]([H:7])([H:8])[O:4][H:10])[O:3][H:9]'
    mapped_mol = utils.load_molecule(mapped_smiles, toolkit=toolkit)

    utils.remove_atom_map(mapped_mol)
    assert utils.has_atom_map(mapped_mol) == False
    assert utils.is_missing_atom_map(mapped_mol) == True

    utils.restore_atom_map(mapped_mol)
    assert utils.has_atom_map(mapped_mol) == True
    assert utils.is_missing_atom_map(mapped_mol) == False

    smiles = 'OCCO'
    mol = utils.load_molecule(smiles, toolkit=toolkit)
    with pytest.warns(UserWarning):
        utils.restore_atom_map(mol)


@pytest.mark.parametrize('toolkit', toolkits_name)
@pytest.mark.parametrize('smiles, canonicalization', [('[H:5][C:1]([H:6])([C:2]([H:7])([H:8])[O:4][H:10])[O:3][H:9]', 'openeye'),
                                                      ('[O:1]([C:3]([C:4]([O:2][H:6])([H:9])[H:10])([H:7])[H:8])[H:5]', 'rdkit')])
def test_is_map_canonical(toolkit, smiles, canonicalization):
    molecule = utils.load_molecule(smiles, toolkit)

    canonical = utils.is_map_canonical(molecule)
    if toolkit == canonicalization:
        assert canonical
    else:
        assert not canonical


@pytest.mark.parametrize('toolkit', toolkits_name)
@pytest.mark.parametrize('smiles', ['COC(C)c1c(Cl)ccc(F)c1Cl',
                                 '[H:3][C:1]([H:4])([H:5])[C:2]([H:6])([H:7])[H:8]',
                         '[H:5][C:1]([H:6])([C:2]([H:7])([H:8])[O:4][H:10])[O:3][H:9]',
                        '[O:1]([C:3]([C:4]([O:2][H:6])([H:9])[H:10])([H:7])[H:8])[H:5]' ])
def test_atom_order_in_mol_copy(toolkit, smiles):
    """Test that atom orders do not change when copying molecule"""
    import copy
    mol = utils.load_molecule(smiles, toolkit=toolkit)
    if not utils.has_explicit_hydrogen(mol):
        mol = utils.add_explicit_hydrogen(mol)
    molcopy = copy.deepcopy(mol)
    for a1, a2 in zip(mol.GetAtoms(), molcopy.GetAtoms()):
        if toolkit == 'openeye':
            assert a1.GetIdx() == a2.GetIdx()
            assert a1.GetName() == a2.GetName()
            assert a1.GetMapIdx() == a2.GetMapIdx()
        if toolkit == 'rdkit':
            assert a1.GetIdx() == a2.GetIdx()
            assert a1.GetAtomMapNum() == a2.GetAtomMapNum()
            assert a1.GetSmarts() == a2.GetSmarts()

@pytest.mark.parametrize('toolkit', toolkits_name)
def test_canonical_label(toolkit):
    """Test canonical label"""
    if toolkit == 'openeye':
        output = '[CH2:3]([CH2:4][OH:2])[OH:1]'
    if toolkit == 'rdkit':
        output = '[OH:1][CH2:3][CH2:4][OH:2]'
    assert utils.to_canonical_label('[O:1]([C:3]([C:4]([O:2][H:6])([H:9])[H:10])([H:7])[H:8])[H:5]',
                                    (0, 1, 2, 3), toolkit) == output
    with pytest.raises(RuntimeError):
        utils.to_canonical_label('OCCO', (0, 1, 2, 3), toolkit)

@using_openeye
def test_int_bond_order():
    """Test bond orders are whole numbers"""
    from openeye import oechem

    hooh = {
        'symbols': ['H', 'O', 'O', 'H'],
        'geometry': [
             1.84719633,  1.47046223,  0.80987166,
             1.3126021,  -0.13023157, -0.0513322,
            -1.31320906,  0.13130216, -0.05020593,
            -1.83756335, -1.48745318,  0.80161212
        ],
        'name': 'HOOH',
        'connectivity': [[0, 1, 1], [1, 2, 1], [2, 3, 1]],
        'molecular_multiplicity': 1
    }
    mol = utils.load_molecule(hooh)
    assert isinstance(mol, oechem.OEMol)

    hooh['connectivity'] = [[0, 1, 1.0], [1, 2, 1.0], [2, 3, 1]]
    mol = utils.load_molecule(hooh)
    assert isinstance(mol, oechem.OEMol)

    hooh['connectivity'] = [[0, 1, 1.0], [1, 2, 1.5], [2, 3, 1.0]]
    with pytest.raises(ValueError):
        utils.load_molecule(hooh)

@using_openeye
def test_n_valence():
    from openeye import oechem
    json_molecule = {'symbols':
                     ['C', 'C', 'C', 'C', 'C', 'C', 'N', 'N', 'N', 'N', 'N', 'H', 'H', 'H', 'H'],
                 'geometry': np.array([11.02088236, 0.30802536, 2.96687012,
                              10.37270642, 2.8383686, 2.75522059,
                              9.32012957, -1.48532476, 2.09948562,
                              8.06346176, 3.48843435, 1.68941515,
                              6.98820713, -0.772898, 1.02801107,
                              5.21186447, -2.73065435, 0.12850138,
                              5.70508328, -5.1797392, 0.28345893,
                              6.45152507, 1.7536658, 0.86549457,
                              2.97820833, -2.31491455, -0.90706852,
                              3.71709131, -6.31357514, -0.68408084,
                              2.05980154, -4.57124733, -1.40784597,
                              12.76887939, -0.24566439, 3.77189345,
                              11.61992628, 4.26322222, 3.39583795,
                              9.76610505, -3.43174262, 2.23743576,
                              7.53811768, 5.41217579, 1.50989122]),
                 'connectivity': [[0, 1, 1],
                                  [0, 2, 2],
                                  [0, 11, 1],
                                  [1, 3, 2],
                                  [1, 12, 1],
                                  [2, 4, 1],
                                  [2, 13, 1],
                                  [3, 7, 1],
                                  [3, 14, 1],
                                  [4, 5, 1],
                                  [4, 7, 2],
                                  [5, 6, 1],
                                  [5, 8, 2],
                                  [6, 9, 1],
                                  [8, 10, 1],
                                  [9, 10, 2]]
                     }
    mol = utils.load_molecule(json_molecule)
    assert utils.has_explicit_hydrogen(mol)
    assert oechem.OEMolToSmiles(mol) == 'c1ccnc(c1)c2[n-]nnn2'


@pytest.mark.parametrize('toolkit', toolkits_name)
def test_add_atom_map(toolkit):
    smiles = 'CCCC'
    mol = utils.load_molecule(smiles, toolkit=toolkit, strict=False)
    mapped_mol = utils.add_atom_map(mol, in_place=False)
    assert utils.has_atom_map(mapped_mol)
    assert not utils.has_atom_map(mol)
    assert not utils.is_missing_atom_map(mapped_mol)
    assert utils.is_missing_atom_map(mol)



"""
Generate canonical identifiers for chemical databases, specifically quantum
chemical data.
"""
from copy import deepcopy
import cmiles
import warnings
from .utils import has_openeye, has_rdkit

if has_openeye:
    import openeye as oe
if has_rdkit:
    import rdkit as rd


def get_molecule_ids(molecule_input, toolkit='openeye', strict=True, **kwargs):
    """
    Generate a dictionary of canonical identifiers

    The `molecule_input` can be either a JSON serialised molecule (see `QCSchema <https://molssi-qc-schema.readthedocs.io/en/latest/index.html#>`_)
    or an isomeric SMILES with all steroechemistry defined.

    Required fields for the QCSchema molecule:

    1. symbols
    2. geometry
    3. connectivity

    Parameters
    ----------
    molecule_input: dict or str
        A JSON serialized QC molecule or an isomeric SMILES
    toolkit: str, optional, default 'openeye'
        toolkit to use for canonicalization. Currently supports `openeye` and `rdkit`
    strict: bool, optional. Default True
        If true, will raise an exception if SMILES is missing explicit H.
    **permute_xyz: bool, optional, default False
        **Only use if input molecule is in QCSchema format**.

        If True, the geometry will be permuted to reflect the canononical
        atom order in the mapped SMILES. ``get_molecule_ids`` will return the permuted QCSchema. ``cmiles`` identifiers will be in
        the `identifiers` field

        If False, the map indices in the mapped SMILES will reflect the order of the atoms in the input QCSchema.

    Returns
    -------
    dict
         If ``permute_xyz=True``, will return permuted qcschema with cmiles identifiers in `identifiers` field.

    """
    # check input and convert to oe or rdkit mol
    molecule = cmiles.utils.load_molecule(molecule_input, toolkit=toolkit, **kwargs)

    # check for map. If map exists, remove. We only want maps generated with cmiles
    if cmiles.utils.has_atom_map(molecule):
        cmiles.utils.remove_atom_map(molecule)

    # check for explicit hydrogen
    if strict and not cmiles.utils.has_explicit_hydrogen(molecule):
        raise ValueError("Input molecule is missing explicit hydrogen")

    # Check for fully defined stereochemsitry
    if not cmiles.utils.has_stereo_defined(molecule):
        raise ValueError("Input SMILES must have full stereochemistry defined")

    molecule_ids = {}
    backend_toolkit = cmiles.utils._set_toolkit(molecule)
    molecule_ids['canonical_smiles'] = backend_toolkit.mol_to_smiles(molecule,
                                                                     isomeric=False,
                                                                     explicit_hydrogen=False,
                                                                     mapped=False)
    molecule_ids['canonical_isomeric_smiles'] = backend_toolkit.mol_to_smiles(molecule,
                                                                              isomeric=True,
                                                                              explicit_hydrogen=False,
                                                                              mapped=False)
    molecule_ids['canonical_explicit_hydrogen_smiles'] = backend_toolkit.mol_to_smiles(molecule,
                                                                                       isomeric=False,
                                                                                       explicit_hydrogen=True,
                                                                                       mapped=False)
    molecule_ids['canonical_isomeric_explicit_hydrogen_smiles'] = backend_toolkit.mol_to_smiles(molecule,
                                                                                                isomeric=True,
                                                                                                explicit_hydrogen=True,
                                                                                                mapped=False)
    molecule_ids['canonical_isomeric_explicit_hydrogen_mapped_smiles'] = backend_toolkit.mol_to_smiles(molecule, isomeric=True,
                                                                                                       explicit_hydrogen=True,
                                                                                                       mapped=True)

    molecule_ids['molecular_formula'] = cmiles.utils.mol_to_hill_molecular_formula(molecule)
    inchi = get_inchi_and_key(molecule)
    if inchi:
        molecule_ids['standard_inchi'] = inchi[0]
        molecule_ids['inchi_key'] = inchi[-1]

    if cmiles.utils.has_rdkit:
        molecule_ids['unique_tautomer_representation'] = standardize_tautomer(molecule_ids['canonical_isomeric_smiles'])

    if cmiles.utils.has_openeye:
        molecule_ids['unique_protomer_representation'] = get_unique_protomer(molecule)

    molecule_ids['provenance'] = 'cmiles_' + cmiles.__version__ + '_{}_'.format(toolkit) + \
                                 backend_toolkit.toolkit.__version__

    try:
        if kwargs['permute_xyz']:
            mapped_smiles = molecule_ids['canonical_isomeric_explicit_hydrogen_mapped_smiles']
            permuted_json_mol = cmiles.utils.permute_qcschema(molecule_input, mapped_smiles, toolkit=toolkit)
            return molecule_ids, permuted_json_mol
        else:
            return molecule_ids

    except KeyError:
        return molecule_ids


def get_inchi_and_key(molecule):
    """
    Generate inchi and inchikey. Uses RDKit which uses the inchi API

    Parameters
    ----------
    molecule: rdkit.Chem.Mol
        If an `oechem.OEMol` is provided, will convert it to an `rdkit.Chem.Mol`

    Returns
    -------
    tuple (inchi, inchi_key)

    """

    # Todo can use the InChI code directly here
    # Make sure molecule is rdkit mol
    if not has_rdkit:
        warnings.warn("rdkit is not installed. cmiles ids will not include inchi and inchikey")
        return
    if not isinstance(molecule, rd.Chem.Mol):
        molecule = rd.Chem.MolFromSmiles(oe.oechem.OEMolToSmiles(molecule))

    inchi = rd.Chem.MolToInchi(molecule)
    inchi_key = rd.Chem.MolToInchiKey(molecule)
    return inchi, inchi_key


def get_iupac(molecule):
    """
    Generate IUPAC name

    Parameters
    ----------
    molecule :
        `oechem.OEMol`

    Returns
    -------
    str:
        iupac name

    Notes
    -----
    Will only be generated if has openeye license

    """
    if not has_openeye:
        raise ImportError("OpenEye is not installed. You can use the canonicalization='rdkit' to use the RDKit backend"
                           "The Conda recipe for cmiles installs rdkit")

    from openeye import oeiupac
    if not oeiupac.OEIUPACIsLicensed():
        raise ImportError("Must have OEIUPAC license!")
    return oeiupac.OECreateIUPACName(molecule)


def get_unique_protomer(molecule):
    """
    Generate unique protomer for all tuatomers and charge states of the moelcule.

    **Requires openeye license**


    Parameters
    ----------
    molecule: oechem.OEMol
        Will convert `rdkit.Chem.Mol` to `oechem.OEMol` if openeye is installed and license is valid

    Returns
    -------
    str
        unique protomer

    """

    molecule = deepcopy(molecule)
    # This only works for OpenEye
    # Todo There might be a way to use different layers of InChI for this purpose.
    # Not all tautomers are recognized as the same by InChI so it won't capture all tautomers.
    # But if we use only the formula and connectivity level we might be able to capture a larger subset.
    #
    if has_openeye:
        from openeye import oequacpac, oechem
    else:
        raise RuntimeError("Must have OpenEye for unique protomer feature")
    if not oechem.OEChemIsLicensed():
        raise ImportError("Must have OEChem license!")
    if not oequacpac.OEQuacPacIsLicensed():
        raise ImportError("Must have OEQuacPac license!")

    if has_rdkit:
        if isinstance(molecule, rd.Chem.rdchem.Mol):
            # convert to openeye molecule
            # Maybe we shouldn't do this.
            smiles = rd.Chem.MolToSmiles(molecule)
            molecule = oechem.OEMol()
            oechem.OESmilesToMol(molecule, smiles)

    molecule_copy = deepcopy(molecule)
    oequacpac.OEGetUniqueProtomer(molecule_copy, molecule)
    return oechem.OEMolToSmiles(molecule_copy)


def standardize_tautomer(iso_can_smi):
    """
    Standardize tautomer to one universal tautomer.

    Parameters
    ----------
    iso_can_smi: str
        isomeric SMILES

    Returns
    -------
    str:
        standardized tautomer

    Notes
    -----
    Does not standardize for ionization states.
    In some cases preforms better than `oequacpac.OEGetUniqueProtomer`.
    See `notebook <https://github.com/openforcefield/cmiles/blob/master/notebooks/Tautomers.ipynb>`_
    """
    if has_rdkit:
        from rdkit.Chem import MolStandardize
    else:
        raise ImportError("Must have rdkit installed to use this function")

    std_tautomer = MolStandardize.canonicalize_tautomer_smiles(iso_can_smi)
    return std_tautomer
